#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tw.addresses: Tracking wallet listaddresses class for the MMGen suite
"""

from ..util import msg, is_int
from ..obj import MMGenListItem, ImmutableAttr, ListItemAttr, TwComment, NonNegativeInt
from ..addr import CoinAddr, MMGenID, MMGenAddrType
from ..amt import CoinAmtChk
from ..color import red, green, yellow
from .view import TwView
from .shared import TwMMGenID

class TwAddresses(TwView):

	hdr_lbl = 'tracking wallet addresses'
	desc = 'address list'
	item_desc = 'address'
	item_desc_pl = 'addresses'
	sort_key = 'twmmid'
	update_widths_on_age_toggle = True
	print_output_types = ('detail',)
	filters = ('showempty', 'showused', 'all_labels')
	showcoinaddrs = True
	showempty = True
	showused = 1 # tristate: 0:no, 1:yes, 2:all
	all_labels = False
	mod_subpath = 'tw.addresses'
	has_age = False
	has_used = False

	prompt_fs_in = [
		'Sort options: [a]mt, [M]mgen addr, [r]everse',
		'Filters: show [E]mpty addrs, show all [L]abels',
		'View/Print: pager [v]iew, [w]ide pager view, [p]rint, r[e]draw{s}',
		'Actions: [q]uit menu, [D]elete addr, add [l]abel, [R]efresh balance:']

	key_mappings = {
		'a':'s_amt',
		'M':'s_twmmid',
		'r':'s_reverse',
		'e':'d_redraw',
		'E':'d_showempty',
		'L':'d_all_labels',
		'l':'i_comment_add',
		'D':'i_addr_delete',
		'v':'a_view',
		'w':'a_view_detail',
		'p':'a_print_detail'}
	extra_key_mappings = {
		'R':'i_balance_refresh'}

	class display_type(TwView.display_type):

		class squeezed(TwView.display_type.squeezed):
			cols = ('num', 'mmid', 'used', 'addr', 'comment', 'amt', 'date')
			fmt_method = 'gen_display'

		class detail(TwView.display_type.detail):
			cols = ('num', 'mmid', 'used', 'addr', 'comment', 'amt', 'block', 'date_time')
			fmt_method = 'gen_display'

	class TwAddress(MMGenListItem):
		valid_attrs = {
			'twmmid',
			'addr',
			'al_id',
			'confs',
			'comment',
			'amt',
			'recvd',
			'is_used',
			'date',
			'skip'}
		invalid_attrs = {'proto'}

		twmmid  = ImmutableAttr(TwMMGenID, include_proto=True) # contains confs,txid(unused),date(unused),al_id
		addr    = ImmutableAttr(CoinAddr, include_proto=True)
		al_id   = ImmutableAttr(str)                          # set to '_' for non-MMGen addresses
		confs   = ImmutableAttr(int, typeconv=False)
		comment = ListItemAttr(TwComment, reassign_ok=True)
		amt     = ImmutableAttr(CoinAmtChk, include_proto=True)
		recvd   = ImmutableAttr(CoinAmtChk, include_proto=True)
		is_used = ImmutableAttr(bool)
		date    = ListItemAttr(int, typeconv=False, reassign_ok=True)
		skip    = ListItemAttr(str, typeconv=False, reassign_ok=True)

		def __init__(self, proto, **kwargs):
			self.__dict__['proto'] = proto
			MMGenListItem.__init__(self, **kwargs)

	@property
	def coinaddr_list(self):
		return [d.addr for d in self.data]

	async def __init__(self, cfg, proto, *, minconf=1, mmgen_addrs='', get_data=False):

		await super().__init__(cfg, proto)

		self.minconf = NonNegativeInt(minconf)
		self.spc_w = 5 + self.has_age + self.has_used # 1 space between cols + 1 leading space in fs
		self.used_w = 4 if self.has_used else 0

		if mmgen_addrs:
			a = mmgen_addrs.rsplit(':', 1)
			if len(a) != 2:
				from ..util import die
				die(1,
					f'{mmgen_addrs}: invalid address list argument ' +
					'(must be in form <seed ID>:[<type>:]<idx list>)')
			from ..addrlist import AddrIdxList
			self.usr_addr_list = [MMGenID(self.proto, f'{a[0]}:{i}') for i in AddrIdxList(fmt_str=a[1])]
		else:
			self.usr_addr_list = []

		if get_data:
			await self.get_data()

	@property
	def no_rpcdata_errmsg(self):
		return 'No addresses {}found!'.format(
			f'with {self.minconf} confirmations ' if self.minconf else '')

	async def get_rpc_data(self):

		self.total = self.proto.coin_amt('0')
		addrs = {}

		used_addrs = self.twctl.used_addrs
		minconf = int(self.minconf)
		block = self.twctl.rpc.get_block_from_minconf(minconf)

		for e in await self.twctl.get_label_addr_pairs():
			bal = await self.twctl.get_balance(e.coinaddr, block=block)
			addrs[e.label.mmid] = {
				'addr':  e.coinaddr,
				'amt':   bal,
				'recvd': bal,         # current bal only, CF btc.tw.addresses.get_rpc_data()
				'is_used': bool(bal) or e.coinaddr in used_addrs,
				'confs': minconf,
				'lbl':   e.label}
			self.total += bal

		return addrs

	async def gen_data(self, rpc_data, lbl_id):
		return (
			self.TwAddress(
					self.proto,
					twmmid  = twmmid,
					addr    = data['addr'],
					al_id   = getattr(twmmid.obj, 'al_id', '_'),
					confs   = data['confs'],
					comment = data['lbl'].comment,
					amt     = data['amt'],
					recvd   = data['recvd'],
					is_used = data['is_used'],
					date    = 0,
					skip    = '')
				for twmmid, data in rpc_data.items()
		)

	def filter_data(self):
		if self.usr_addr_list:
			return (d for d in self.data if d.twmmid.obj in self.usr_addr_list)
		else:
			return (d for d in self.data if
				(self.all_labels and d.comment) or
				(self.showused == 2 and d.is_used) or
				(not (d.is_used and not self.showused) and (d.amt or self.showempty))
			)

	def get_column_widths(self, data, *, wide, interactive):

		return self.compute_column_widths(
			widths = { # fixed cols
				'num':  max(2, len(str(len(data)))+1),
				'mmid': max(len(d.twmmid.disp) for d in data),
				'used': self.used_w,
				'amt':  self.amt_widths['amt'],
				'date': self.age_w if self.has_age else 0,
				'block': self.age_col_params['block'][0] if wide and self.has_age else 0,
				'date_time': self.age_col_params['date_time'][0] if wide and self.has_age else 0,
				'spc': self.spc_w,
			},
			maxws = { # expandable cols
				'addr':    max(len(d.addr) for d in data) if self.showcoinaddrs else 0,
				'comment': max(d.comment.screen_width for d in data),
			},
			minws = {
				'addr': 12 if self.showcoinaddrs else 0,
				'comment': len('Comment'),
			},
			maxws_nice = {'addr': 18},
			wide = wide,
			interactive = interactive,
		)

	def squeezed_col_hdr(self, cw, fs, color):
		return fs.format(
			n  = '',
			m  = 'MMGenID',
			u  = 'Used',
			a  = 'Address',
			c  = 'Comment',
			A  = 'Balance',
			d  = self.age_hdr)

	def detail_col_hdr(self, cw, fs, color):
		return fs.format(
			n  = '',
			m  = 'MMGenID',
			u  = 'Used',
			a  = 'Address',
			c  = 'Comment',
			A  = 'Balance',
			b  = 'Block',
			D  = 'Date/Time')

	def squeezed_format_line(self, n, d, cw, fs, color, yes, no):
		return fs.format(
			n = str(n) + ')',
			m = d.twmmid.fmt(cw.mmid, color=color),
			u = yes if d.is_used else no,
			a = d.addr.fmt(self.addr_view_pref, cw.addr, color=color),
			c = d.comment.fmt2(cw.comment, color=color, nullrepl='-'),
			A = d.amt.fmt(cw.iwidth, color=color, prec=self.disp_prec),
			d = self.age_disp(d, self.age_fmt)
		)

	def detail_format_line(self, n, d, cw, fs, color, yes, no):
		return fs.format(
			n = str(n) + ')',
			m = d.twmmid.fmt(cw.mmid, color=color),
			u = yes if d.is_used else no,
			a = d.addr.fmt(self.addr_view_pref, cw.addr, color=color),
			c = d.comment.fmt2(cw.comment, color=color, nullrepl='-'),
			A = d.amt.fmt(cw.iwidth, color=color, prec=self.disp_prec),
			b = self.age_disp(d, 'block'),
			D = self.age_disp(d, 'date_time'))

	def gen_display(self, data, cw, fs, color, fmt_method):

		yes, no = (red('Yes '), green('No  ')) if color else ('Yes ', 'No  ')
		id_save = data[0].al_id

		for n, d in enumerate(data, 1):
			if id_save != d.al_id:
				id_save = d.al_id
				yield ''.ljust(self.term_width)
			yield fmt_method(n, d, cw, fs, color, yes, no)

	async def set_dates(self, addrs):
		if not self.dates_set:
			bc = self.rpc.blockcount + 1
			caddrs = [addr for addr in addrs if addr.confs]
			hashes = await self.rpc.gathered_call('getblockhash', [(n,) for n in [bc - a.confs for a in caddrs]])
			dates = [d['time'] for d in await self.rpc.gathered_call('getblockheader', [(h,) for h in hashes])]
			for idx, addr in enumerate(caddrs):
				addr.date = dates[idx]
			self.dates_set = True

	sort_disp = {
		'age': 'AddrListID+Age',
		'amt': 'AddrListID+Amt',
		'twmmid': 'MMGenID',
	}

	sort_funcs = {
		'age': lambda d: '{}_{}_{}'.format(
			d.al_id,
			# Hack, but OK for the foreseeable future:
			('{:>012}'.format(1_000_000_000 - d.confs) if d.confs else '_'),
			d.twmmid.sort_key),
		'amt': lambda d: f'{d.al_id}_{d.amt}',
		'twmmid': lambda d: d.twmmid.sort_key,
	}

	@property
	def dump_fn_pfx(self):
		return 'listaddresses' + (f'-minconf-{self.minconf}' if self.minconf else '')

	@property
	def sid_ranges(self):

		def gen_sid_ranges():

			from collections import namedtuple
			sid_range = namedtuple('sid_range', ['bot', 'top'])

			sid_save = None
			bot = None

			for n, e in enumerate(self.data):
				if e.twmmid.type == 'mmgen':
					if e.twmmid.obj.sid != sid_save:
						if sid_save:
							yield (sid_save, sid_range(bot, n-1))
						sid_save = e.twmmid.obj.sid
						bot = n
				else:
					break
			else:
				n += 1

			if sid_save:
				yield (sid_save, sid_range(bot, n-1))

		assert self.sort_key == 'twmmid'
		assert self.reverse is False

		if not hasattr(self, '_sid_ranges'):
			self._sid_ranges = dict(gen_sid_ranges())

		return self._sid_ranges

	def is_used(self, coinaddr):
		for e in self.data:
			if e.addr == coinaddr:
				return e.is_used
		return None # addr not in tracking wallet

	def get_change_address(self, al_id, *, bot=None, top=None, exclude=None, desc=None):
		"""
		Get lowest-indexed unused address in tracking wallet for requested AddrListID.
		Return values on failure:
		    None:  no addresses in wallet with requested AddrListID
		    False: no unused addresses in wallet with requested AddrListID
		"""

		def get_start(bot, top):
			"""
			bisecting algorithm to find first entry with requested al_id

			Since 'btc' > 'F' and pre_target sorts below the first twmmid of the al_id
			stringwise, we can just search on raw twmmids.
			"""
			pre_target = al_id + ':0'
			n = top >> 1

			while True:

				if bot == top:
					return bot if data[bot].al_id == al_id else None

				if data[n].twmmid < pre_target:
					bot = n + 1
				else:
					top = n

				n = (top + bot) >> 1

		assert self.sort_key == 'twmmid'
		assert self.reverse is False

		data = self.data
		start = get_start(
			bot = 0             if bot is None else bot,
			top = len(data) - 1 if top is None else top)

		if start is not None:
			for d in data[start:]:
				if d.al_id == al_id:
					if (
							not d.is_used
							and not d.twmmid in exclude
							and (self.cfg.autochg_ignore_labels or not d.comment)
						):
						if d.comment:
							msg('{} {} {} {}{}'.format(
								yellow('WARNING: address'),
								d.twmmid.hl(),
								yellow('has a label,'),
								d.comment.hl2(encl='‘’'),
								yellow(f',\n  but allowing it for {desc} anyway by user request')
							))
						return d
				else:
					break
			return False

	def get_change_address_by_addrtype(self, mmtype, *, exclude, desc):
		"""
		Find the lowest-indexed change addresses in tracking wallet of given address type,
		present them in a menu and return a single change address chosen by the user.

		If mmtype is None, search all preferred_mmtypes in tracking wallet

		Return values on failure:
		    None:  no addresses in wallet of requested address type
		    False: no unused addresses in wallet of requested address type
		"""

		def choose_address(addrs):

			def format_line(n, d):
				return '{a:3}) {b}{c}'.format(
					a = n,
					b = d.twmmid.hl(),
					c = yellow(' <== has a label!') if d.comment else ''
				)

			prompt = '\nChoose a {desc}:\n\n{items}\n\nEnter a number> '.format(
				desc = desc,
				items = '\n'.join(format_line(n, d) for n, d in enumerate(addrs, 1)))

			from ..ui import line_input
			while True:
				res = line_input(self.cfg, prompt)
				if is_int(res) and 0 < int(res) <= len(addrs):
					return addrs[int(res)-1]
				msg(f'{res}: invalid entry')

		def get_addr(mmtype):
			return [self.get_change_address(
				f'{sid}:{mmtype}', bot=r.bot, top=r.top, exclude=exclude, desc=desc)
					for sid, r in self.sid_ranges.items()]

		assert isinstance(mmtype, (type(None), MMGenAddrType))

		if mmtype:
			res = get_addr(mmtype)
		else:
			have_used = False
			for mmtype in self.proto.preferred_mmtypes:
				res = get_addr(mmtype)
				if any(res):
					break
				if False in res:
					have_used = True
			else:
				return False if have_used else None

		if any(res):
			res = list(filter(None, res))
			if len(res) == 1:
				return res[0]
			else:
				return choose_address(res)
		elif False in res:
			return False

	class display_action(TwView.display_action):

		def d_showempty(self, parent):
			parent.showempty = not parent.showempty

		def d_showused(self, parent):
			parent.showused = (parent.showused + 1) % 3

		def d_all_labels(self, parent):
			parent.all_labels = not parent.all_labels



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
rpc: RPC library for the MMGen Project
"""

import importlib

from ..util import die, fmt, fmt_list
from ..obj import NonNegativeInt

async def rpc_init(
		cfg,
		proto                 = None,
		*,
		backend               = None,
		daemon                = None,
		ignore_daemon_version = False,
		ignore_wallet         = False):

	proto = proto or cfg._proto

	if not 'rpc_init' in proto.mmcaps:
		die(1, f'rpc_init() not supported for {proto.name} protocol!')

	if proto.rpc_type == 'remote':
		return getattr(importlib.import_module(
			f'mmgen.proto.{proto.base_proto_coin.lower()}.rpc.remote'),
				proto.base_proto + 'RemoteRPCClient')(cfg=cfg, proto=proto)

	from ..daemon import CoinDaemon

	rpc = await getattr(importlib.import_module(
			f'mmgen.proto.{proto.base_proto_coin.lower()}.rpc.local'),
				proto.base_proto + 'RPCClient')(
		cfg           = cfg,
		proto         = proto,
		daemon        = daemon or CoinDaemon(cfg, proto=proto, test_suite=cfg.test_suite),
		backend       = backend or cfg.rpc_backend,
		ignore_wallet = ignore_wallet)

	if rpc.daemon_version > rpc.daemon.coind_version:
		rpc.handle_unsupported_daemon_version(
			proto.name,
			ignore_daemon_version or proto.ignore_daemon_version or cfg.ignore_daemon_version)

	if rpc.chain not in proto.chain_names:
		die('RPCChainMismatch', '\n' + fmt(f"""
			Protocol:           {proto.cls_name}
			Valid chain names:  {fmt_list(proto.chain_names, fmt='bare')}
			RPC client chain:   {rpc.chain}
			""", indent='  ').rstrip())

	rpc.blockcount = NonNegativeInt(rpc.blockcount)

	return rpc



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
devtools: Developer tools for the MMGen suite
"""

import sys

color_funcs = {
	name: lambda s, n=n: f'\033[{n};1m{s}\033[0m'
		for name, n in (
			('red',    31),
			('green',  32),
			('yellow', 33),
			('blue',   34),
			('purple', 35))
}

def pfmt(*args, color=None):
	import pprint
	ret = pprint.PrettyPrinter(indent=4, width=116).pformat(
		args if len(args) > 1 else '' if not args else args[0])
	return color_funcs[color](ret) if color else ret

def pmsg(*args, color=None):
	sys.stderr.write(pfmt(*args, color=color) + '\n')

def pmsg_r(*args, color=None):
	sys.stderr.write(pfmt(*args, color=color))

def pdie(*args, exit_val=1):
	pmsg(*args, color='red' if exit_val else None)
	sys.exit(exit_val)

def pexit(*args):
	pdie(*args, exit_val=0)

def Pmsg(*args, color=None):
	sys.stdout.write(pfmt(*args, color=color) + '\n')

def Pdie(*args, exit_val=1):
	Pmsg(*args, color=('yellow' if exit_val == 1 else 'red' if exit_val else None))
	sys.exit(exit_val)

def Pexit(*args):
	Pdie(*args, exit_val=0)

def print_stack_trace(message=None, fh_list=[], nl='\n', sep='\n  ', trim=4):
	if not fh_list:
		import os
		fh_list.append(open(f'devtools.trace.{os.getpid()}', 'w'))
		nl = ''
	res = get_stack_trace(message, nl, sep, trim)
	sys.stderr.write(res)
	fh_list[0].write(res)

def get_stack_trace(message=None, nl='\n', sep='\n  ', trim=3):

	import os, re, traceback

	tb = [t for t in traceback.extract_stack() if t.filename[:1] != '<']
	fs = '{}:{}: in {}:\n    {}'
	out = [
		fs.format(
			re.sub(r'^\./', '', os.path.relpath(t.filename)),
			t.lineno,
			(t.name+'()' if t.name[-1] != '>' else t.name),
			t.line or '(none)')
		for t in (tb[:-trim] if trim else tb)]

	return f'{nl}STACK TRACE {message or "[unnamed]"}:{sep}{sep.join(out)}\n'

def print_diff(*args, **kwargs):
	sys.stderr.write(get_diff(*args, **kwargs))

def get_diff(a, b, a_fn='', b_fn='', from_json=True):

	if from_json:
		import json
		a = json.dumps(json.loads(a), indent=4)
		b = json.dumps(json.loads(b), indent=4)

	from difflib import unified_diff
	# chunk headers have trailing newlines, hence the rstrip()
	return '  DIFF:\n    {}\n'.format(
		'\n    '.join(a.rstrip('\n') for a in unified_diff(
			a.split('\n'),
			b.split('\n'),
			a_fn,
			b_fn)))

def print_ndiff(*args, **kwargs):
	sys.stderr.write(get_ndiff(*args, **kwargs))

def get_ndiff(a, b):
	from difflib import ndiff
	return list(ndiff(
		a.split('\n'),
		b.split('\n')))

class MMGenObjectMethods: # mixin class for MMGenObject

	# Pretty-print an MMGenObject instance, recursing into sub-objects - WIP
	def pmsg(self, color=None):
		sys.stdout.write('\n'+self.pfmt(color=color))

	def pdie(self, exit_val=1):
		self.pmsg(color='red' if exit_val else None)
		sys.exit(exit_val)

	def pexit(self):
		self.pdie(exit_val=0)

	def pfmt(self, lvl=0, id_list=[], color=None):

		from decimal import Decimal
		scalars = (str, int, float, Decimal)

		def isDict(obj):
			return isinstance(obj, dict)
		def isList(obj):
			return isinstance(obj, list)
		def isScalar(obj):
			return isinstance(obj, scalars)

		def do_list(out, e, *, lvl=0, is_dict=False):
			out.append('\n')
			for i in e:
				el = i if not is_dict else e[i]
				if is_dict:
#					out.append('{s}{:<{l}}'.format(i, s=' '*(4*lvl+8), l=10, l2=8*(lvl+1)+8))
					out.append('{s1}{i}{s2}'.format(
						i  = i,
						s1 = ' ' * (4*lvl+8),
						s2 = ' ' * 10))
				if hasattr(el, 'pfmt'):
					out.append('{:>{l}}{}'.format(
						'',
						el.pfmt(lvl=lvl+1, id_list=id_list+[id(self)]),
						l = (lvl+1)*8))
				elif isinstance(el, scalars):
					if isList(e):
						out.append('{:>{l}}{!r:16}\n'.format('', el, l=lvl*8))
					else:
						out.append(f' {el!r}')
				elif isList(el) or isDict(el):
					indent = 1 if is_dict else lvl*8+4
					out.append('{:>{l}}{:16}'.format('', f'<{type(el).__name__}>', l=indent))
					if isList(el) and isinstance(el[0], scalars):
						out.append('\n')
					do_list(out, el, lvl=lvl+1, is_dict=isDict(el))
				else:
					out.append('{:>{l}}{:16} {!r}\n'.format('', f'<{type(el).__name__}>', el, l=(lvl*8)+8))
				out.append('\n')

			if not e:
				out.append(f'{e!r}\n')

		out = [f'<{type(self).__name__}>{" "+repr(self) if isScalar(self) else ""}\n']

		if id(self) in id_list:
			return out[-1].rstrip() + ' [RECURSION]\n'
		if isList(self) or isDict(self):
			do_list(out, self, lvl=lvl, is_dict=isDict(self))

		for k in self.__dict__:
			e = getattr(self, k)
			if isList(e) or isDict(e):
				out.append('{:>{l}}{:<10} {:16}'.format('', k, f'<{type(e).__name__}>', l=(lvl*8)+4))
				do_list(out, e, lvl=lvl, is_dict=isDict(e))
			elif hasattr(e, 'pfmt') and callable(e.pfmt) and not isinstance(e, type):
				out.append('{:>{l}}{:10} {}'.format(
					'',
					k,
					e.pfmt(lvl=lvl+1, id_list=id_list+[id(self)]),
					l = (lvl*8)+4))
			else:
				out.append('{:>{l}}{:<10} {:16} {}\n'.format(
					'',
					k,
					f'<{type(e).__name__}>',
					repr(e),
					l=(lvl*8)+4))

		import re
		ret = re.sub('\n+', '\n', ''.join(out))
		return color_funcs[color](ret) if color else ret



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.include.common: Shared routines and data for the MMGen test suites
"""

import sys, os, re, atexit
from subprocess import run, PIPE, DEVNULL
from pathlib import Path
from collections import namedtuple

from mmgen.cfg import gv
from mmgen.color import yellow, green, orange
from mmgen.util import msg, msg_r, Msg, Msg_r, gmsg, die, suf, fmt_list
from mmgen.fileutil import write_data_to_file, get_data_from_file

def noop(*args, **kwargs):
	pass

def set_globals(cfg):
	"""
	make `cfg`, `qmsg`, `vmsg`, etc. available as globals to scripts by setting
	the module attr
	"""
	import test.include.common as this
	this.cfg = cfg

	if cfg.quiet:
		this.qmsg = this.qmsg_r = noop
	else:
		this.qmsg = msg
		this.qmsg_r = msg_r

	if cfg.verbose:
		this.vmsg = msg
		this.vmsg_r = msg_r
		this.Vmsg = Msg
		this.Vmsg_r = Msg_r
	else:
		this.vmsg = this.vmsg_r = this.Vmsg = this.Vmsg_r = noop

	this.dmsg = msg if cfg.debug else noop

def strip_ansi_escapes(s):
	return re.sub('\x1b' + r'\[[;0-9]+?m', '', s)

cmdtest_py_log_fn = 'cmdtest.py.log'
cmdtest_py_error_fn = 'cmdtest.py.err'
parity_dev_amt = 1606938044258990275541962092341162602522202993782792835301376
ascii_uc   = ''.join(map(chr, list(range(65, 91))))   # 26 chars
ascii_lc   = ''.join(map(chr, list(range(97, 123))))  # 26 chars
lat_accent = ''.join(map(chr, list(range(192, 383)))) # 191 chars, L, S
ru_uc = ''.join(map(chr, list(range(1040, 1072)))) # 32 chars
gr_uc = ''.join(map(chr, list(range(913, 930)) + list(range(931, 940)))) # 26 chars (930 is ctrl char)
gr_uc_w_ctrl = ''.join(map(chr, list(range(913, 940)))) # 27 chars, L, C
lat_cyr_gr = lat_accent[:130:5] + ru_uc + gr_uc # 84 chars
ascii_cyr_gr = ascii_uc + ru_uc + gr_uc # 84 chars

utf8_text      = '[α-$ample UTF-8 text-ω]' * 10   # 230 chars, L, N, P, S, Z
utf8_combining = '[α-$ámple UTF-8 téxt-ω]' * 10   # L, N, P, S, Z, M
utf8_ctrl      = '[α-$ample\nUTF-8\ntext-ω]' * 10 # L, N, P, S, Z, C

text_jp = '必要なのは、信用ではなく暗号化された証明に基づく電子取引システムであり、これにより希望する二者が信用できる第三者機関を介さずに直接取引できるよう' # 72 chars ('W'ide)
text_zh = '所以，我們非常需要這樣一種電子支付系統，它基於密碼學原理而不基於信用，使得任何達成一致的雙方，能夠直接進行支付，從而不需要協力廠商仲介的參與。。' # 72 chars ('F'ull + 'W'ide)

sample_text = 'The Times 03/Jan/2009 Chancellor on brink of second bailout for banks'
sample_mn = {
	'mmgen': { # 'able': 0, 'youth': 1625, 'after' == 'afternoon'[:5]
		'mn': 'able cast forgive master funny gaze after afternoon million paint moral youth',
		'hex': '0005685ab4e94cbe3b228cf92112bc5f',
	},
	'bip39': { # len('sun') < uniq_ss_len
		'mn': 'vessel ladder alter error federal sibling chat ability sun glass valve picture',
		'hex': 'f30f8c1da665478f49b001d94c5fc452',
	},
	'xmrseed': {
		'mn': 'viewpoint donuts ardent template unveil agile meant unafraid urgent athlete rustled mime azure jaded hawk baby jagged haystack baby jagged haystack ramped oncoming point template',
		'hex': 'e8164dda6d42bd1e261a3406b2038dcbddadbeefdeadbeefdeadbeefdeadbe0f',
	},
}

ref_kafile_pass = 'kafile password'
ref_kafile_hash_preset = '1'

proto_cmds = (
	'addrimport',
	'autosign',
	'msg',
	'regtest',
	'tool',
	'txbump',
	'txcreate',
	'txdo',
	'txsend',
	'txsign',
	'xmrwallet',
)

def getrand(n):
	if cfg.test_suite_deterministic:
		from mmgen.test import fake_urandom
		return fake_urandom(n)
	else:
		return os.urandom(n)

def getrandnum(n):
	return int(getrand(n).hex(), 16)

def getrandhex(n):
	return getrand(n).hex()

def getrandnum_range(nbytes, rn_max):
	while True:
		rn = int(getrand(nbytes).hex(), 16)
		if rn < rn_max:
			return rn

def getrandstr(num_chars, no_space=False):
	n, m = (94, 33) if no_space else (95, 32)
	return ''.join(chr(i % n + m) for i in list(getrand(num_chars)))

# Windows uses non-UTF8 encodings in filesystem, so use raw bytes here
def cleandir(d, do_msg=False):
	d_enc = d.encode()

	try:
		files = os.listdir(d_enc)
	except:
		return None

	if files:
		from shutil import rmtree
		if do_msg:
			gmsg(f'Cleaning directory {d!r}')
		for f in files:
			try:
				os.unlink(os.path.join(d_enc, f))
			except:
				rmtree(os.path.join(d_enc, f), ignore_errors=True)

	return files

def mk_tmpdir(d):
	try:
		os.makedirs(d, mode=0o755, exist_ok=True)
	except OSError as e:
		if e.errno != 17:
			raise
	else:
		vmsg(f'Created directory {d!r}')

def clean(cfgs, tmpdir_ids=None, extra_dirs=[]):

	def clean_tmpdirs():
		cfg_tmpdirs = {k:cfgs[k]['tmpdir'] for k in cfgs}
		for d in map(str, sorted(map(int, (tmpdir_ids or cfg_tmpdirs)))):
			if d in cfg_tmpdirs:
				if cleandir(cfg_tmpdirs[d]):
					yield d
			else:
				die(1, f'{d}: invalid directory number')

	def clean_extra_dirs():
		for d in extra_dirs:
			if os.path.exists(d):
				if cleandir(d):
					yield os.path.relpath(d)

	for clean_func, list_fmt in (
				(clean_tmpdirs, 'no_quotes'),
				(clean_extra_dirs, 'dfl')
			):
		if cleaned := list(clean_func()):
			iqmsg(green('Cleaned director{} {}'.format(
				suf(cleaned, 'ies'),
				fmt_list(cleaned, fmt=list_fmt)
			)))

	for d in extra_dirs:
		if (os.path.exists(d) or os.path.islink(d)) and not os.path.isdir(d):
			print(f'Removing non-directory ‘{d}’')
			os.unlink(d)

def get_tmpfile(cfg, fn):
	return os.path.join(cfg['tmpdir'], fn)

def write_to_file(fn, data, binary=False):
	write_data_to_file(
		cfg,
		fn,
		data,
		quiet = True,
		no_stdout = True,
		binary = binary,
		ignore_opt_outdir = True)

def write_to_tmpfile(cfg, fn, data, binary=False):
	write_to_file(os.path.join(cfg['tmpdir'], fn), data=data, binary=binary)

def read_from_file(fn, binary=False):
	return get_data_from_file(cfg, fn, quiet=True, binary=binary)

def read_from_tmpfile(cfg, fn, binary=False):
	return read_from_file(os.path.join(cfg['tmpdir'], fn), binary=binary)

def joinpath(*args, **kwargs):
	return os.path.join(*args, **kwargs)

def ok(text='OK'):
	if cfg.profile:
		return
	if cfg.verbose or cfg.exact_output:
		gmsg(text)
	else:
		msg(f' {text}')

def cmp_or_die(s, t, desc=None):
	if s != t:
		die('TestSuiteFatalException',
			(f'For {desc}:\n' if desc else '') +
			f'ERROR: recoded data:\n{t!r}\ndiffers from original data:\n{s!r}'
		)

def chk_equal(a, b):
	assert a == b, f'equality test failed: {a} != {b}'

def init_coverage():
	coverdir = os.path.join('test', 'trace')
	acc_file = os.path.join('test', 'trace.acc')
	try:
		os.mkdir(coverdir, 0o755)
	except:
		pass
	return coverdir, acc_file

def silence():
	if not (cfg.verbose or cfg.exact_output):
		gv.stdout = gv.stderr = open(os.devnull, 'w')

def end_silence():
	if not (cfg.verbose or cfg.exact_output):
		gv.stdout.close()
		gv.stdout = sys.stdout
		gv.stderr = sys.stderr

def omsg(s):
	sys.stderr.write(s + '\n')
def omsg_r(s):
	sys.stderr.write(s)
	sys.stderr.flush()

def imsg(s):
	if cfg.verbose or cfg.exact_output:
		omsg(s)
def imsg_r(s):
	if cfg.verbose or cfg.exact_output:
		omsg_r(s)

def iqmsg(s):
	if not cfg.quiet:
		omsg(s)
def iqmsg_r(s):
	if not cfg.quiet:
		omsg_r(s)

def oqmsg(s):
	if not (cfg.verbose or cfg.exact_output):
		omsg(s)
def oqmsg_r(s):
	if not (cfg.verbose or cfg.exact_output):
		omsg_r(s)

def end_msg(t):
	omsg(green(
		'All requested tests finished OK' +
		('' if cfg.test_suite_deterministic else f', elapsed time: {t//60:02d}:{t%60:02d}')
	))

def start_test_daemons(*network_ids, remove_datadir=False):
	if not cfg.no_daemon_autostart:
		return test_daemons_ops(*network_ids, op='start', remove_datadir=remove_datadir)

def stop_test_daemons(*network_ids, force=False, remove_datadir=False):
	if force or not cfg.no_daemon_stop:
		return test_daemons_ops(*network_ids, op='stop', remove_datadir=remove_datadir)

def restart_test_daemons(*network_ids, remove_datadir=False):
	if not stop_test_daemons(*network_ids, remove_datadir=remove_datadir):
		return False
	return start_test_daemons(*network_ids, remove_datadir=remove_datadir)

def test_daemons_ops(*network_ids, op, remove_datadir=False):
	if not cfg.no_daemon_autostart:
		from mmgen.daemon import CoinDaemon
		silent = not (cfg.verbose or cfg.exact_output)
		ret = False
		for network_id in network_ids:
			d = CoinDaemon(cfg, network_id=network_id, test_suite=True)
			if remove_datadir:
				d.wait = True
				d.stop(silent=True)
				d.remove_datadir()
			ret = d.cmd(op, silent=silent)
		return ret

tested_solc_ver = '0.8.26'

def check_solc_ver():
	cmd = 'python3 scripts/create-token.py --check-solc-version'
	try:
		cp = run(cmd.split(), check=False, stdout=PIPE)
	except Exception as e:
		die(4, f'Unable to execute {cmd!r}: {e}')
	res = cp.stdout.decode().strip()
	if cp.returncode == 0:
		omsg(
			orange(f'Found supported solc version {res}') if res == tested_solc_ver else
			yellow(f'WARNING: solc version ({res}) does not match tested version ({tested_solc_ver})')
		)
		return True
	else:
		omsg(yellow('Warning: Solidity compiler (solc) could not be executed or has unsupported version'))
		omsg(res)
		return False

def get_ethkey():
	cmdnames = ('ethkey', 'openethereum-ethkey')
	for cmdname in cmdnames:
		try:
			cp = run([cmdname, '--help'], stdout=PIPE, text=True)
		except:
			pass
		else:
			if 'Parity' in cp.stdout:
				return cmdname
	return None

def do_run(cmd, check=True):
	return run(cmd, stdout=PIPE, stderr=DEVNULL, check=check)

def test_exec(cmd):
	try:
		do_run(cmd.split())
		return True
	except Exception:
		return False

def in_nix_environment():
	for path in os.getenv('PATH').split(':'):
		if os.path.realpath(path).startswith('/nix/store'):
			return True
	return False

def make_burn_addr(proto, mmtype='compressed', hexdata=None):
	from mmgen.tool.coin import tool_cmd
	return tool_cmd(
		cfg     = cfg,
		cmdname = 'pubhash2addr',
		proto   = proto,
		mmtype  = mmtype).pubhash2addr(hexdata or '00'*20)

def create_addrpairs(proto, mmtype, num):
	ap = namedtuple('addrpair', ['wif', 'addr'])
	from mmgen.tool.coin import tool_cmd
	n = 123456789123456789
	return [ap(*tool_cmd(
		cfg     = cfg,
		cmdname = 'privhex2pair',
		proto   = proto,
		mmtype  = mmtype).privhex2pair(f'{n+m:064x}'))
			for m in range(num)]

def VirtBlockDevice(img_path, size):
	if sys.platform == 'linux':
		return VirtBlockDeviceLinux(img_path, size)
	elif sys.platform == 'darwin':
		return VirtBlockDeviceMacOS(img_path, size)

class VirtBlockDeviceBase:

	@property
	def dev(self):
		res = self._get_associations()
		if len(res) < 1:
			die(2, f'No device associated with {self.img_path}')
		elif len(res) > 1:
			die(2, f'More than one device associated with {self.img_path}')
		return res[0]

	def try_detach(self):
		try:
			dev = self.dev
		except:
			pass
		else:
			self.do_detach(dev, check=False)

	def create(self, silent=False):
		for dev in self._get_associations():
			if not silent:
				imsg(f'Detaching associated device {dev}')
			self.do_detach(dev)
		self.img_path.unlink(missing_ok=True)
		if not silent:
			imsg(f'Creating block device image file {self.img_path}')
		self.do_create(self.size, self.img_path)
		atexit.register(self.try_detach)

	def attach(self, dev_mode=None, silent=False):
		if res := self._get_associations():
			die(2, f'Device{suf(res)} {fmt_list(res, fmt="barest")} already associated with {self.img_path}')
		dev = self.get_new_dev()
		if dev_mode:
			self.dev_mode_orig = '0{:o}'.format(os.stat(dev).st_mode & 0xfff)
			if not silent:
				imsg(f'Changing permissions on device {dev} to {dev_mode!r}')
			do_run(['sudo', 'chmod', dev_mode, dev])
		if not silent:
			imsg(f'Attaching {dev or self.img_path!r}')
		self.do_attach(self.img_path, dev)

	def detach(self, silent=False):
		dev = self.dev
		if not silent:
			imsg(f'Detaching {dev!r}')
		self.do_detach(dev)
		if hasattr(self, 'dev_mode_orig'):
			if not silent:
				imsg(f'Resetting permissions on device {dev} to {self.dev_mode_orig!r}')
			do_run(['sudo', 'chmod', self.dev_mode_orig, dev])
			delattr(self, 'dev_mode_orig')

	def __del__(self):
		self.try_detach()

class VirtBlockDeviceLinux(VirtBlockDeviceBase):

	def __init__(self, img_path, size):
		self.img_path = Path(img_path).resolve()
		self.size = size

	def _get_associations(self):
		cmd = ['sudo', 'losetup', '-n', '-O', 'NAME', '-j', str(self.img_path)]
		return do_run(cmd).stdout.decode().splitlines()

	def get_new_dev(self):
		return do_run(['sudo', 'losetup', '-f']).stdout.decode().strip()

	def do_create(self, size, path):
		do_run(['truncate', f'--size={size}', str(path)])

	def do_attach(self, path, dev):
		do_run(['sudo', 'losetup', dev, str(path)])

	def do_detach(self, dev, check=True):
		do_run(['sudo', 'losetup', '-d', dev], check=check)

class VirtBlockDeviceMacOS(VirtBlockDeviceBase):

	def __init__(self, img_path, size):
		self.img_path = Path(img_path + '.dmg').resolve()
		self.size = size

	def _get_associations(self):
		cp = run(['hdiutil', 'info'], stdout=PIPE, stderr=PIPE, text=True, check=False)
		if cp.returncode == 0:
			lines = cp.stdout.splitlines()
			out = [re.sub('.* ', '', s.strip()) for s in lines if re.match(r'image-path|/dev/', s)]
			def gen_pairs():
				for n in range(len(out) // 2):
					yield(out[n*2], out[(n*2)+1])
			return [dev for path, dev in gen_pairs() if path == str(self.img_path)]
		else:
			return []

	def get_new_dev(self):
		return None

	def do_create(self, size, path):
		do_run(['hdiutil', 'create', '-size', size, '-layout', 'NONE', str(path)])

	def do_attach(self, path, dev=None):
		do_run(['hdiutil', 'attach', '-nomount', str(path)])

	def do_detach(self, dev, check=True):
		do_run(['hdiutil', 'detach', dev], check=check)



#!/usr/bin/env python3

"""
test/start-coin-daemons.py: Start daemons for the MMGen test suite
"""

try:
	from include.coin_daemon_control import main
except ImportError:
	from test.include.coin_daemon_control import main

from mmgen.main import launch
launch(func=main)



from .autosign_orig import *

class overlay_fake_Autosign:

	def init_fixup(self):
		if pfx := self.cfg.test_suite_root_pfx:
			subdir = pfx + '/' + ('online' if self.cfg.online else 'offline')
			for k in ('mountpoint', 'shm_dir', 'wallet_dir'):
				orig_path = str(getattr(self, k))
				setattr(self, k, Path(subdir + orig_path.removeprefix(subdir)))
			# mount --type=fuse-ext2 --options=rw+ ### current fuse-ext2 (0.4 29) is buggy - can’t use
			import sys
			if sys.platform == 'linux':
				self.dev_label = 'MMGEN_TS_ONLINE' if self.cfg.online else 'MMGEN_TS_OFFLINE'
				self.mount_cmd  = f'sudo mount LABEL={self.dev_label} {self.mountpoint}'
				self.umount_cmd = f'sudo umount {self.mountpoint}'
				self.linux_blkid_cmd = 'sudo blkid -s LABEL -o value'

Autosign.dev_label          = 'MMGEN_TS_TX' # autosign_live only (Linux)
Autosign.linux_mount_subdir = 'mmgen_ts_autosign'
Autosign.macOS_ramdisk_name = 'TestAutosignRamDisk'
Autosign.init_fixup         = overlay_fake_Autosign.init_fixup



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
help.swaptxcreate: swaptxcreate and swaptxdo help notes for the MMGen Wallet suite
"""

def help(proto, cfg):
	return """
This script is similar in operation to ‘mmgen-txcreate’, only with additional
steps.  Users are advised to first familiarize themselves with the use of that
script before attempting to perform a swap with this one.

The tracking wallets of both the send and receive coins must be available when
the script is invoked.  If the two coin daemons are running on different hosts
than the script, or with non-standard ports, coin-specific RPC options may be
required (see EXAMPLES below).

The swap protocol’s quote server on the Internet must be reachable either
directly or via the SOCKS5 proxy specified with the --proxy option. To improve
privacy, it’s recommended to proxy requests to the quote server via Tor or
some other anonymity network.

The resulting transaction file is saved, signed, sent, and optionally bumped,
exactly the same way as one created with ‘mmgen-txcreate’.  Autosign with
automount is likewise supported via the --autosign option.

The command line must contain at minimum a send coin (COIN1) and receive coin
(COIN2) symbol.  Currently supported coins are BTC, LTC, BCH and ETH.  All
other arguments are optional.  If AMT is specified, the specified value of
send coin will be swapped and the rest returned to a change address in the
originating tracking wallet.  Otherwise, the entire value of the interactively
selected inputs will be swapped.

By default, the change (if applicable) and destination addresses are chosen
automatically by finding the lowest-indexed unused addresses of the preferred
address types in the send and receive tracking wallets.  For Bitcoin and
forks, types ‘B’, ‘S’ and ‘C’ (see ADDRESS TYPES below) are searched in that
order for unused addresses.  Note that sending to an unused address may be
undesirable for Ethereum, where address (i.e. account) reuse is the norm.  In
that case, the user should specify a destination address on the command line.

If the wallet contains eligible unused addresses with multiple Seed IDs, the
user will be presented with a list of the lowest-indexed addresses of
preferred type for each Seed ID and prompted to choose from among them.

Change and destination addresses may also be specified manually with the
CHG_ADDR and ADDR arguments.  These may be given as full MMGen IDs or in the
form ADDRTYPE_CODE or SEED_ID:ADDRTYPE_CODE (see EXAMPLES below and the
‘mmgen-txcreate’ help screen for details).  For Ethereum, the CHG_ADDR
argument is not supported.

While discouraged, sending change or swapping to non-wallet addresses is also
supported, in which case the signing script (‘mmgen-txsign’ or ‘mmgen-
autosign’, as applicable) must be invoked with the --allow-non-wallet-swap
option.

Rather than specifying a transaction fee on the command line, it’s advisable
to start with the fee suggested by the swap protocol quote server (the script
does this automatically) and then adjust the fee interactively if desired.

When choosing a fee, bear in mind that the longer the transaction remains
unconfirmed, the greater the risk that the vault address will expire, leading
to loss of funds.  It’s therefore recommended to learn how to create, sign and
send replacement transactions with ‘mmgen-txbump’ before performing a swap
with this script.  When bumping a stuck swap transaction, the safest option
is to create a replacement transaction with one output that returns funds back
to the originating tracking wallet, thus aborting the swap, rather than one
that merely increases the fee (see EXAMPLES below).

Before broadcasting the transaction, it’s a good idea to double-check the
vault address on a block explorer such as thorchain.net or runescan.io.

The MMGen Node Tools suite contains two useful tools to help with fine-tuning
transaction fees, ‘mmnode-feeview’ and ‘mmnode-blocks-info’, in addition to
‘mmnode-ticker’, which can be used to calculate the current cross-rate between
the asset pair of a swap, as well as the total receive value in terms of send
value.


                                TRADE LIMIT

A target value for the swap may be set, known as the “trade limit”.  If
this target cannot be met, the network will refund the user’s coins, minus
transaction fees (note that the refund goes to the address associated with the
transaction’s first input, leading to coin reuse).  Since under certain
circumstances large amounts of slippage can occur, resulting in significant
losses, setting a trade limit is highly recommended.

The target may be given as either an absolute coin amount or percentage value.
In the latter case, it’s interpreted as the percentage below the “expected
amount out” returned by the swap quote server.  Zero or negative percentage
values are also accepted, but are likely to result in your coins being
refunded.

The trade limit is rounded to four digits of precision in order to reduce
transaction size.
"""



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test/objtest.py: Test MMGen data objects
"""

import os, re

try:
	from include import test_init
except ImportError:
	from test.include import test_init

# for objtest, violate MMGen Project best practices and allow use of the dev tools
# in production code:
from mmgen.devtools import pmsg
if not os.getenv('MMGEN_DEVTOOLS'):
	from mmgen.devinit import init_dev
	init_dev()

from mmgen.cfg import Config
from mmgen.util import msg, msg_r, gmsg, capfirst, die
from mmgen.color import red, yellow, blue, green, orange, purple, gray, nocolor
from mmgen.obj import get_obj

opts_data = {
	'sets': [('super_silent', True, 'silent', True)],
	'text': {
		'desc': 'Test MMGen data objects',
		'usage':'[options] [object]',
		'options': """
-h, --help         Print this help message
--, --longhelp     Print help message for long (global) options
-g, --getobj       Instantiate objects with get_obj() wrapper
-q, --quiet        Produce quieter output
-s, --silent       Silence output of tested objects
-S, --super-silent Silence all output except for errors
-v, --verbose      Produce more verbose output
"""
	}
}

cfg = Config(opts_data=opts_data)

if cfg.verbose:
	from mmgen.objmethods import MMGenObject

from test.include.common import set_globals
set_globals(cfg)

def run_test(mod, test, arg, input_data, arg1, exc_name):
	arg_copy = arg
	kwargs = {}
	ret_chk = arg
	ret_idx = None
	if input_data == 'good' and isinstance(arg, tuple):
		arg, ret_chk = arg
	if isinstance(arg, dict): # pass one arg + kwargs to constructor
		arg_copy = arg.copy()
		if 'arg' in arg:
			args = [arg['arg']]
			ret_chk = args[0]
			del arg['arg']
		else:
			args = []
			ret_chk = list(arg.values())[0] # assume only one key present
		if 'ret' in arg:
			ret_chk = arg['ret']
			del arg['ret']
			del arg_copy['ret']
		if 'exc_name' in arg:
			exc_name = arg['exc_name']
			del arg['exc_name']
			del arg_copy['exc_name']
		if 'ret_idx' in arg:
			ret_idx = arg['ret_idx']
			del arg['ret_idx']
			del arg_copy['ret_idx']
		kwargs.update(arg)
	elif isinstance(arg, tuple):
		args = arg
	else:
		args = [arg]

	if cfg.getobj:
		if args:
			assert len(args) == 1, 'objtest_chk1: only one positional arg is allowed'
			kwargs.update( { arg1: args[0] } )
		if cfg.silent:
			kwargs.update( { 'silent': True } )

	try:
		if not cfg.super_silent:
			arg_disp = repr(arg_copy[0] if isinstance(arg_copy, tuple) else arg_copy)
			if cfg.test_suite_deterministic and isinstance(arg_copy, dict):
				arg_disp = re.sub(r'object at 0x[0-9a-f]+', 'object at [SCRUBBED]', arg_disp)
			msg_r((green if input_data=='good' else orange)(f'{arg_disp+":":<22}'))
		cls = getattr(mod, test)

		if cfg.getobj:
			ret = get_obj(getattr(mod, test), **kwargs)
		else:
			ret = cls(*args, **kwargs)

		bad_ret = [] if issubclass(cls, list) else None

		if isinstance(ret_chk, str):
			ret_chk = ret_chk.encode()

		if isinstance(ret, str):
			ret = ret.encode()

		if cfg.getobj:
			if input_data == 'bad':
				assert ret is False, 'non-False return on bad input data'
		else:
			if (cfg.silent and input_data=='bad' and ret!=bad_ret) or (not cfg.silent and input_data=='bad'):
				raise UserWarning(f"Non-'None' return value {ret!r} with bad input data")

		if cfg.silent and input_data=='good' and ret==bad_ret:
			raise UserWarning("'None' returned with good input data")

		if input_data=='good':
			if ret_idx:
				ret_chk = arg[list(arg.keys())[ret_idx]].encode()
			if ret != ret_chk and repr(ret) != repr(ret_chk):
				raise UserWarning(f"Return value ({ret!r}) doesn't match expected value ({ret_chk!r})")

		if cfg.super_silent:
			return

		if cfg.getobj and (not cfg.silent and input_data == 'bad'):
			pass
		else:
			try:
				ret_disp = ret.decode()
			except:
				ret_disp = ret
			msg(f'==> {ret_disp!r}')

		if cfg.verbose and issubclass(cls, MMGenObject):
			ret.pmsg() if hasattr(ret, 'pmsg') else pmsg(ret)

	except UserWarning as e:
		msg(f'==> {ret!r}')
		die(2, red(str(e)))
	except Exception as e:
		if input_data == 'good':
			raise ValueError(f'Error on good input data: {e}') from e
		if not type(e).__name__ == exc_name:
			msg(f'Incorrect exception: expected {exc_name} but got {type(e).__name__}')
			raise
		if cfg.super_silent:
			pass
		elif cfg.silent:
			msg(f'==> {exc_name}')
		else:
			msg( yellow(f' {exc_name}:') + str(e) )
	except SystemExit as e:
		if input_data == 'good':
			raise ValueError('Error on good input data') from e
		if cfg.verbose:
			msg(f'exitval: {e.code}')

def do_loop():
	import importlib
	modname = f'test.objtest_d.{proto.coin.lower()}_{proto.network}'
	mod = importlib.import_module(modname)
	test_data = getattr(mod, 'tests')
	gmsg(f'Running data object tests for {proto.coin} {proto.network}')

	clr = None
	utests = cfg._args
	for test in test_data:
		arg1 = test_data[test].get('arg1')
		if utests and test not in utests:
			continue
		nl = ('\n', '')[bool(cfg.super_silent) or clr is None]
		clr = (blue, nocolor)[bool(cfg.super_silent)]

		if cfg.getobj and arg1 is None:
			msg(gray(f'{nl}Skipping {test}'))
			continue

		msg(clr(f'{nl}Testing {test}'))

		for k in ('bad', 'good'):
			if not cfg.super_silent:
				msg(purple(capfirst(k)+' input:'))
			for arg in test_data[test][k]:
				run_test(
					mod,
					test,
					arg,
					input_data = k,
					arg1       = arg1,
					exc_name   = test_data[test].get('exc_name') or ('ObjectInitError', 'None')[k=='good'])

proto = cfg._proto

if __name__ == '__main__':
	do_loop()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
proto.eth.addrdata: Ethereum TwAddrData classes
"""

from ...addrdata import TwAddrDataWithStore

class EthereumTwAddrData(TwAddrDataWithStore):
	pass

class EthereumTokenTwAddrData(EthereumTwAddrData):
	pass



from .tx_proxy_orig import *

class overlay_fake_EtherscanTxProxyClient:
	network_proto = 'http'
	host = 'localhost:28800'
	verify = False

EtherscanTxProxyClient.network_proto = overlay_fake_EtherscanTxProxyClient.network_proto
EtherscanTxProxyClient.host = overlay_fake_EtherscanTxProxyClient.host
EtherscanTxProxyClient.verify = overlay_fake_EtherscanTxProxyClient.verify



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_d.ltc_mainnet: LTC mainnet test vectors for MMGen data objects
"""

from decimal import Decimal

from mmgen.amt import LTCAmt
from mmgen.addr import CoinAddr
from mmgen.key import WifKey, PrivKey
from mmgen.protocol import init_proto

from .common import r16, r32
from ..include.common import cfg

proto = init_proto(cfg, 'ltc', need_amt=True)

tests = {
	'LTCAmt': {
		'bad': ('-3.2', '0.123456789', 123, '123L', '88000000', 80999999.12345678),
		'good': (('80999999.12345678', Decimal('80999999.12345678')),)
	},
	'CoinAddr': {
		'bad': (
			{'addr': 1,   'proto': proto},
			{'addr': 'x', 'proto': proto},
			{'addr': 'я', 'proto': proto},
		),
		'good': (
			{'addr': 'LXYx4j8PDGE8GEwDFnEQhcLyHFGsRxSJwt', 'proto': proto},
			{'addr': 'MEnuCzUGHaQx9fK5WYvLwR1NK4SAo8HmSr', 'proto': proto},
		),
	},
	'WifKey': {
		'bad': (
			{'proto': proto, 'wif': 1},
			{'proto': proto, 'wif': []},
			{'proto': proto, 'wif': '\0'},
			{'proto': proto, 'wif': '\1'},
			{'proto': proto, 'wif': 'я'},
			{'proto': proto, 'wif': 'g'},
			{'proto': proto, 'wif': 'gg'},
			{'proto': proto, 'wif': 'FF'},
			{'proto': proto, 'wif': 'f00'},
			{'proto': proto, 'wif': r16.hex()},
			{'proto': proto, 'wif': '2MspvWFjBbkv2wzQGqhxJUYPCk3Y2jMaxLN'},
		),
		'good': (
			{'proto': proto, 'wif': '6udBAGS6B9RfGyvEQDkVDsWy3Kqv9eTULqtEfVkJtTJyHdLvojw',  'ret_idx': 1},
			{'proto': proto, 'wif': 'T7kCSp5E71jzV2zEJW4q5qU1SMB5CSz8D9VByxMBkamv1uM3Jjca', 'ret_idx': 1},
		)
	},
	'PrivKey': {
		'bad': (
			{'proto': proto, 'wif': 1},
			{'proto': proto, 'wif': '1'},
			{'proto': proto, 'wif': 'cMsqcmDYZP1LdKgqRh9L4ZRU9br28yvdmTPwW2YQwVSN9aQiMAoR'},
			{'proto': proto, 's': r32, 'wif': '6ufJhtQQiRYA3w2QvDuXNXuLgPFp15i3HR1Wp8An2mx1JnhhJAh'},
			{'proto': proto, 'pubkey_type': 'std'},
			{'proto': proto, 's': r32},
			{'proto': proto, 's': r32, 'compressed': 'yes'},
			{'proto': proto, 's': r32, 'compressed': 'yes', 'pubkey_type': 'std'},
			{'proto': proto, 's': r32, 'compressed': True, 'pubkey_type': 'nonstd'},
			{'proto': proto, 's': r32+b'x', 'compressed': True, 'pubkey_type': 'std'}
		),
		'good': (
			{'proto': proto, 'wif': '6ufJhtQQiRYA3w2QvDuXNXuLgPFp15i3HR1Wp8An2mx1JnhhJAh',
			'ret': bytes.fromhex('470a974ffca9fca1299b706b09142077bea3acbab6d6480b87dbba79d5fd279b')},
			{'proto': proto, 'wif': 'T41Fm7J3mtZLKYPMCLVSFARz4QF8nvSDhLAfW97Ds56Zm9hRJgn8',
			'ret': bytes.fromhex('1c6feab55a4c3b4ad1823d4ecacd1565c64228c01828cf44fb4db1e2d82c3d56')},
			{'proto': proto, 's': r32, 'compressed': False, 'pubkey_type': 'std', 'ret': r32},
			{'proto': proto, 's': r32, 'compressed': True, 'pubkey_type': 'std', 'ret': r32}
		)
	},
}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.unenc: unencrypted wallet base class
"""

from ..color import blue, yellow
from ..util import msg, msg_r, capfirst, is_int
from .base import wallet

class wallet(wallet):

	def _decrypt_retry(self):
		pass

	def _encrypt(self):
		pass

	def _filename(self):
		s = self.seed
		return '{}[{}].{}'.format(s.fn_stem, s.bitlen, self.ext)

	def _print_seed_type(self):
		msg('{} {}'.format(
			blue(f'{capfirst(self.base_type or self.type)} type:'),
			yellow(self.mn_type)
		))

	def _choose_seedlen(self, ok_lens):

		from ..term import get_char
		def choose_len():
			prompt = self.choose_seedlen_prompt
			while True:
				r = get_char('\r'+prompt)
				if is_int(r) and 1 <= int(r) <= len(ok_lens):
					break
			msg_r(('\r', '\n')[self.cfg.test_suite] + ' '*len(prompt) + '\r')
			return ok_lens[int(r)-1]

		while True:
			usr_len = choose_len()
			prompt = self.choose_seedlen_confirm.format(usr_len)
			from ..ui import keypress_confirm
			if keypress_confirm(
					self.cfg,
					prompt,
					default_yes = True,
					no_nl       = not self.cfg.test_suite):
				return usr_len



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.relay: Monero wallet ops for the MMGen Suite
"""

import time
from pathlib import Path

from ...util import msg, msg_r, gmsg, ymsg, die
from ...ui import keypress_confirm
from ...proto.xmr.rpc import MoneroRPCClient

from ..file.tx import MoneroMMGenTX

from . import OpBase

class OpRelay(OpBase):
	opts = ('tx_relay_daemon',)

	def __init__(self, cfg, uarg_tuple):

		super().__init__(cfg, uarg_tuple)

		self.mount_removable_device()

		self.tx = MoneroMMGenTX.Signed(self.cfg, Path(self.uargs.infile))

		if self.cfg.tx_relay_daemon:
			m = self.parse_tx_relay_opt()
			host, port = m[1].split(':')
			proxy = m[2]
			md = None
		else:
			from ...daemon import CoinDaemon
			md = CoinDaemon(self.cfg, network_id='xmr', test_suite=self.cfg.test_suite)
			host, port = ('localhost', md.rpc_port)
			proxy = None

		self.dc = MoneroRPCClient(
			cfg    = self.cfg,
			proto  = self.proto,
			daemon = md,
			host   = host,
			port   = int(port),
			user   = None,
			passwd = None,
			test_connection = host == 'localhost', # avoid extra connections if relay is a public node
			proxy  = proxy)

	async def main(self):
		msg('\n' + self.tx.get_info(indent='    '))

		if self.cfg.tx_relay_daemon:
			self.display_tx_relay_info(indent='    ')

		keypress_confirm(self.cfg, 'Relay transaction?', do_exit=True)

		if self.cfg.tx_relay_daemon:
			msg_r('Relaying transaction to remote daemon, please be patient...')
			t_start = time.time()

		res = self.dc.call_raw('send_raw_transaction', tx_as_hex=self.tx.data.blob)
		if res['status'] == 'OK':
			if res['not_relayed']:
				msg('not relayed')
				ymsg('Transaction not relayed')
			else:
				msg('success')
			if self.cfg.tx_relay_daemon:
				from ...util2 import format_elapsed_hr
				msg(f'Relay time: {format_elapsed_hr(t_start, rel_now=False, show_secs=True)}')
			gmsg('OK')
			return True
		else:
			die('RPCFailure', repr(res))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.new_swap: THORChain new swap transaction class
"""

from ....tx.new_swap import NewSwap as TxNewSwap

from ...vm.tx.new_swap import VmNewSwap

from .new import New

class NewSwap(VmNewSwap, New, TxNewSwap):
	desc = 'RUNE swap transaction'

	def update_vault_addr(self, c, *, addr='inbound_address'):
		pass



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.completed: Bitcoin completed transaction class
"""

from ....tx import completed as TxBase
from ....obj import HexStr
from ....util import msg, die
from .base import Base, decodeScriptPubKey

class Completed(Base, TxBase.Completed):
	fn_fee_unit = 'satoshi'

	def get_swap_memo_maybe(self):
		if o := self.data_output:
			return o.data

	# check signature and witness data
	def check_sigs(self): # return True if sigs found, False otherwise; raise exception on error
		txins = self.deserialized.txins
		has_ss = any(ti['scriptSig'] for ti in txins)
		has_witness = any('witness' in ti and ti['witness'] for ti in txins)
		if not (has_ss or has_witness):
			return False
		fs = "Hex TX has {} scriptSig but input is of type '{}'!"
		for n, ti in enumerate(txins):
			mmti = self.inputs[n]
			if ti['scriptSig'] == '' or (len(ti['scriptSig']) == 46 and # native P2WPKH or P2SH-P2WPKH
					ti['scriptSig'][:6] == '16' + self.proto.witness_vernum_hex + '14'):
				assert 'witness' in ti, 'missing witness'
				assert isinstance(ti['witness'], list) and len(ti['witness']) == 2, 'malformed witness'
				assert len(ti['witness'][1]) == 66, 'incorrect witness pubkey length'
				assert mmti.mmtype == ('S', 'B')[ti['scriptSig']==''], fs.format('witness-type', mmti.mmtype)
			else: # non-witness
				assert mmti.mmtype not in ('S', 'B'), fs.format('signature in', mmti.mmtype)
				assert not 'witness' in ti, 'non-witness input has witness'
				# sig_size 72 (DER format), pubkey_size 'compressed':33, 'uncompressed':65
				assert (200 < len(ti['scriptSig']) < 300), 'malformed scriptSig' # VERY rough check
		return True

	def check_pubkey_scripts(self):
		for n, i in enumerate(self.inputs, 1):
			ds = decodeScriptPubKey(self.proto, i.scriptPubKey)
			if ds.addr != i.addr:
				if ds.addr_fmt != i.addr.addr_fmt:
					m = 'Address format of scriptPubKey ({}) does not match that of address ({}) in input #{}'
					msg(m.format(ds.addr_fmt, i.addr.addr_fmt, n))
				m = 'ERROR: Address and scriptPubKey of transaction input #{} do not match!'
				die(3, (m+'\n  {:23}{}'*3).format(
					n,
					'address:',               i.addr,
					'scriptPubKey:',          i.scriptPubKey,
					'scriptPubKey->address:', ds.addr))

	def is_replaceable(self):
		return self.inputs[0].sequence == self.proto.max_int - 2

	@property
	def send_amt(self):
		return self.sum_outputs(
			exclude = None if len(self.nondata_outputs) == 1 else self.chg_idx
		)

	def check_txfile_hex_data(self):
		self.serialized = HexStr(self.serialized)

	def parse_txfile_serialized_data(self):
		pass

	@property
	def fee(self):
		return self.sum_inputs() - self.sum_outputs()

	@property
	def change(self):
		return self.sum_outputs() - self.send_amt

	def get_serialized_locktime(self):
		return int(bytes.fromhex(self.serialized[-8:])[::-1].hex(), 16)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
addr: MMGen address-related types
"""

from collections import namedtuple

from .objmethods import HiliteStr, InitErrors, MMGenObject
from .obj import ImmutableAttr, MMGenIdx, get_obj
from .seed import SeedID
from . import color as color_mod

ati = namedtuple('addrtype_info',
	['name', 'pubkey_type', 'compressed', 'gen_method', 'addr_fmt', 'wif_label', 'extra_attrs', 'desc'])

class MMGenAddrType(HiliteStr, InitErrors, MMGenObject):
	width = 1
	trunc_ok = False
	color = 'blue'

	name        = ImmutableAttr(str)
	pubkey_type = ImmutableAttr(str)
	compressed  = ImmutableAttr(bool, set_none_ok=True)
	gen_method  = ImmutableAttr(str, set_none_ok=True)
	addr_fmt    = ImmutableAttr(str, set_none_ok=True)
	wif_label   = ImmutableAttr(str, set_none_ok=True)
	extra_attrs = ImmutableAttr(tuple, set_none_ok=True)
	desc        = ImmutableAttr(str)

	pkh_fmts = ('p2pkh', 'bech32', 'ethereum')
	mmtypes = {
		'L': ati('legacy',    'std', False,'p2pkh',   'p2pkh',   'wif', (), 'Legacy uncompressed address'),
		'C': ati('compressed','std', True, 'p2pkh',   'p2pkh',   'wif', (), 'Compressed P2PKH address'),
		'S': ati('segwit',    'std', True, 'segwit',  'p2sh',    'wif', (), 'Segwit P2SH-P2WPKH address'),
		'B': ati('bech32',    'std', True, 'bech32',  'bech32',  'wif', (), 'Native Segwit (Bech32) address'),
		'X': ati('bech32x',   'std', True, 'p2pkh',   'bech32',  'wif', (), 'Cross-chain Bech32 address'),
		'E': ati('ethereum',  'std', False,'ethereum','p2pkh',   'privkey', ('wallet_passwd',),'Ethereum address'),
		'Z': ati('zcash_z','zcash_z',False,'zcash_z', 'zcash_z', 'wif',     ('viewkey',),      'Zcash z-address'),
		'M': ati('monero', 'monero', False,'monero',  'monero',  'spendkey',('viewkey','wallet_passwd'),'Monero address'),
	}
	def __new__(cls, proto, id_str, *, errmsg=None):
		if isinstance(id_str, cls):
			return id_str
		try:
			id_str = id_str.replace('-', '_')
			for k, v in cls.mmtypes.items():
				if id_str in (k, v.name):
					if id_str == v.name:
						id_str = k
					me = str.__new__(cls, id_str)
					for k in v._fields:
						setattr(me, k, getattr(v, k))
					if me not in proto.mmtypes + ('P',):
						raise ValueError(f'{me.name!r}: invalid address type for {proto.name} protocol')
					me.proto = proto
					return me
			raise ValueError(f'{id_str}: unrecognized address type for protocol {proto.name}')
		except Exception as e:
			return cls.init_fail(
				e,
				f"{errmsg or ''}{id_str!r}: invalid value for {cls.__name__} ({e!s})",
				preformat = True)

	@classmethod
	def get_names(cls):
		return [v.name for v in cls.mmtypes.values()]

def is_mmgen_addrtype(proto, id_str):
	return get_obj(MMGenAddrType, proto=proto, id_str=id_str, silent=True, return_bool=True)

class MMGenPasswordType(MMGenAddrType):
	mmtypes = {
		'P': ati('password', 'password', None, None, None, None, None, 'Password generated from MMGen seed')
	}

class AddrIdx(MMGenIdx):
	max_digits = 7

def is_addr_idx(s):
	return get_obj(AddrIdx, n=s, silent=True, return_bool=True)

class AddrListID(HiliteStr, InitErrors, MMGenObject):
	width = 10
	trunc_ok = False
	color = 'yellow'
	def __new__(cls, *, sid=None, mmtype=None, proto=None, id_str=None):
		try:
			if id_str:
				a, b = id_str.split(':')
				sid = SeedID(sid=a)
				try:
					mmtype = MMGenAddrType(proto=proto, id_str=b)
				except:
					mmtype = MMGenPasswordType(proto=proto, id_str=b)
			else:
				assert isinstance(sid, SeedID), f'{sid!r} not a SeedID instance'
				if not isinstance(mmtype, (MMGenAddrType, MMGenPasswordType)):
					raise ValueError(f'{mmtype!r}: not an instance of MMGenAddrType or MMGenPasswordType')
			me = str.__new__(cls, sid+':'+mmtype)
			me.sid = sid
			me.mmtype = mmtype
			return me
		except Exception as e:
			return cls.init_fail(e, f'sid={sid}, mmtype={mmtype}, id_str={id_str}')

def is_addrlist_id(proto, s):
	return get_obj(AddrListID, proto=proto, id_str=s, silent=True, return_bool=True)

class MMGenID(HiliteStr, InitErrors, MMGenObject):
	color = 'orange'
	width = 0
	trunc_ok = False
	def __new__(cls, proto, id_str):
		try:
			ss = str(id_str).split(':')
			assert len(ss) in (2, 3), 'not 2 or 3 colon-separated items'
			t = proto.addr_type((ss[1], proto.dfl_mmtype)[len(ss)==2])
			me = str.__new__(cls, f'{ss[0]}:{t}:{ss[-1]}')
			me.sid = SeedID(sid=ss[0])
			me.idx = AddrIdx(ss[-1])
			me.mmtype = t
			assert t in proto.mmtypes, f'{t}: invalid address type for {proto.cls_name}'
			me.al_id = str.__new__(AddrListID, me.sid+':'+me.mmtype) # checks already done
			me.sort_key = f'{me.sid}:{me.mmtype}:{me.idx:0{me.idx.max_digits}}'
			me.proto = proto
			return me
		except Exception as e:
			return cls.init_fail(e, id_str)

def is_mmgen_id(proto, s):
	return get_obj(MMGenID, proto=proto, id_str=s, silent=True, return_bool=True)

class CoinAddr(HiliteStr, InitErrors, MMGenObject):
	color = 'cyan'
	hex_width = 40
	width = 1
	trunc_ok = False

	def __new__(cls, proto, addr):
		if isinstance(addr, cls):
			return addr
		try:
			ap = proto.decode_addr(addr)
			assert ap, f'coin address {addr!r} could not be parsed'
			if hasattr(ap, 'addr'):
				me = str.__new__(cls, ap.addr)
				me.views = ap.views
				me.view_pref = ap.view_pref
			else:
				me = str.__new__(cls, addr)
				me.views = [addr]
				me.view_pref = 0
			me.addr_fmt = ap.fmt
			me.bytes = ap.bytes
			me.ver_bytes = ap.ver_bytes
			me.proto = proto
			return me
		except Exception as e:
			return cls.init_fail(e, addr, objname=f'{proto.name} {proto.cls_name} address')

	@property
	def parsed(self):
		if not hasattr(self, '_parsed'):
			self._parsed = self.proto.parse_addr(self.ver_bytes, self.bytes, self.addr_fmt)
		return self._parsed

	# reimplement some HiliteStr methods:
	@classmethod
	def fmtc(cls, s, width, /, *, color=False):
		return super().fmtc(s[:width-2]+'..' if len(s) > width else s, width, color=color)

	def fmt(self, view_pref, width, /, *, color=False):
		s = self.views[view_pref]
		return super().fmtc(f'{s[:width-2]}..' if len(s) > width else s, width, color=color)

	def hl(self, view_pref, /, *, color=True):
		return getattr(color_mod, self.color)(self.views[view_pref]) if color else self.views[view_pref]

def is_coin_addr(proto, s):
	return get_obj(CoinAddr, proto=proto, addr=s, silent=True, return_bool=True)

class ContractAddr(CoinAddr):
	color = 'blue'

def ViewKey(proto, viewkey_str):
	return proto.viewkey(viewkey_str)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.coin: Cryptocoin routines for the 'mmgen-tool' utility
"""

from collections import namedtuple
generator_data = namedtuple('generator_data', ['kg', 'ag'])

from .common import tool_cmd_base

from ..key import PrivKey
from ..addr import CoinAddr, MMGenAddrType
from ..addrgen import KeyGenerator, AddrGenerator

class tool_cmd(tool_cmd_base):
	"""
	cryptocoin key/address utilities

	May require use of the '--coin', '--type' and/or '--testnet' options

	Examples:
	  mmgen-tool --coin=ltc --type=bech32 wif2addr <wif key>
	  mmgen-tool --coin=zec --type=zcash_z randpair
	"""

	need_proto = True
	need_addrtype = True

	def _init_generators(self):
		return generator_data(
			kg = KeyGenerator(self.cfg, self.proto, self.mmtype.pubkey_type),
			ag = AddrGenerator(self.cfg, self.proto, self.mmtype),
		)

	def randwif(self):
		"generate a random private key in WIF format"
		from ..crypto import Crypto
		return PrivKey(
			self.proto,
			Crypto(self.cfg).get_random(32),
			pubkey_type = self.mmtype.pubkey_type,
			compressed  = self.mmtype.compressed).wif

	def privhex2pair(self, privhex: 'sstr'):
		"generate a wifkey/address pair from the provided hexadecimal key"
		gd = self._init_generators()
		privkey = PrivKey(
			self.proto,
			bytes.fromhex(privhex),
			pubkey_type = self.mmtype.pubkey_type,
			compressed  = self.mmtype.compressed)
		return (
			privkey.wif,
			gd.ag.to_addr(gd.kg.gen_data(privkey)))

	def randpair(self):
		"generate a random wifkey/address pair"
		from ..crypto import Crypto
		return self.privhex2pair(Crypto(self.cfg).get_random(32).hex())

	def wif2hex(self, wifkey: 'sstr'):
		"convert a private key from WIF to hexadecimal format"
		return PrivKey(
			self.proto,
			wif = wifkey).hex()

	def hex2wif(self, privhex: 'sstr'):
		"convert a private key from hexadecimal to WIF format"
		return PrivKey(
			self.proto,
			bytes.fromhex(privhex),
			pubkey_type = self.mmtype.pubkey_type,
			compressed  = self.mmtype.compressed).wif

	def wif2addr(self, wifkey: 'sstr'):
		"generate a coin address from a key in WIF format"
		gd = self._init_generators()
		privkey = PrivKey(
			self.proto,
			wif = wifkey)
		return gd.ag.to_addr(gd.kg.gen_data(privkey))

	def wif2redeem_script(self, wifkey: 'sstr'): # new
		"convert a WIF private key to a Segwit P2SH-P2WPKH redeem script"
		assert self.mmtype.name == 'segwit', 'This command is meaningful only for --type=segwit'
		gd = self._init_generators()
		privkey = PrivKey(
			self.proto,
			wif = wifkey)
		return gd.ag.to_segwit_redeem_script(gd.kg.gen_data(privkey))

	def wif2segwit_pair(self, wifkey: 'sstr'):
		"generate a Segwit P2SH-P2WPKH redeem script and address from a WIF private key"
		assert self.mmtype.name == 'segwit', 'This command is meaningful only for --type=segwit'
		gd = self._init_generators()
		data = gd.kg.gen_data(PrivKey(
			self.proto,
			wif = wifkey))
		return (
			gd.ag.to_segwit_redeem_script(data),
			gd.ag.to_addr(data))

	def _privhex2out(self, privhex: 'sstr', *, output_pubhex=False):
		gd = self._init_generators()
		pk = PrivKey(
			self.proto,
			bytes.fromhex(privhex),
			compressed  = self.mmtype.compressed,
			pubkey_type = self.mmtype.pubkey_type)
		data = gd.kg.gen_data(pk)
		return data.pubkey.hex() if output_pubhex else gd.ag.to_addr(data)

	def privhex2addr(self, privhex: 'sstr'):
		"generate a coin address from raw hexadecimal private key data"
		return self._privhex2out(privhex)

	def privhex2pubhex(self, privhex: 'sstr'): # new
		"generate a hexadecimal public key from raw hexadecimal private key data"
		return self._privhex2out(privhex, output_pubhex=True)

	def pubhex2addr(self, pubkeyhex: 'sstr'):
		"convert a hexadecimal pubkey to an address"
		if self.proto.base_proto == 'Ethereum' and len(pubkeyhex) == 128: # support raw ETH pubkeys
			pubkeyhex = '04' + pubkeyhex
		from ..keygen import keygen_public_data
		ag = AddrGenerator(self.cfg, self.proto, self.mmtype)
		return ag.to_addr(keygen_public_data(
			pubkey        = bytes.fromhex(pubkeyhex),
			viewkey_bytes = None,
			pubkey_type   = self.mmtype.pubkey_type,
			compressed    = self.mmtype.compressed,
		))

	def pubhex2redeem_script(self, pubkeyhex: 'sstr'): # new
		"convert a hexadecimal pubkey to a Segwit P2SH-P2WPKH redeem script"
		assert self.mmtype.name == 'segwit', 'This command is meaningful only for --type=segwit'
		from ..proto.btc.common import hash160
		return self.proto.pubhash2redeem_script(hash160(bytes.fromhex(pubkeyhex))).hex()

	def redeem_script2addr(self, redeem_script_hex: 'sstr'): # new
		"convert a Segwit P2SH-P2WPKH redeem script to an address"
		assert self.mmtype.name == 'segwit', 'This command is meaningful only for --type=segwit'
		assert redeem_script_hex[:4] == '0014', f'{redeem_script_hex!r}: invalid redeem script'
		assert len(redeem_script_hex) == 44, f'{len(redeem_script_hex)//2} bytes: invalid redeem script length'
		from ..proto.btc.common import hash160
		return self.proto.pubhash2addr(hash160(bytes.fromhex(redeem_script_hex)), 'p2sh')

	def pubhash2addr(self, pubhashhex: 'sstr'):
		"convert public key hash to address"
		pubhash = bytes.fromhex(pubhashhex)
		if self.mmtype.name == 'segwit':
			return self.proto.pubhash2segwitaddr(pubhash)
		elif self.mmtype.name == 'bech32':
			return self.proto.pubhash2bech32addr(pubhash)
		elif self.mmtype.name == 'bech32x':
			return self.proto.encode_addr_bech32x(pubhash)
		else:
			return self.proto.pubhash2addr(pubhash, self.mmtype.addr_fmt)

	def addr2pubhash(self, addr: 'sstr'):
		"convert coin address to public key hash"
		ap = self.proto.decode_addr(addr)
		assert ap, f'coin address {addr!r} could not be parsed'
		if ap.fmt not in MMGenAddrType.pkh_fmts:
			from ..util import die
			die(2, f'{ap.fmt} addresses cannot be converted to pubhash')
		return ap.bytes.hex()

	def addr2scriptpubkey(self, addr: 'sstr'):
		"convert coin address to scriptPubKey"
		from ..proto.btc.tx.base import addr2scriptPubKey
		return addr2scriptPubKey(self.proto, CoinAddr(self.proto, addr))

	def scriptpubkey2addr(self, hexstr: 'sstr'):
		"convert scriptPubKey to coin address"
		from ..proto.btc.tx.base import decodeScriptPubKey
		return decodeScriptPubKey(self.proto, hexstr).addr

	def eth_checksummed_addr(self, addr: 'sstr'):
		"create a checksummed Ethereum address"
		from ..protocol import init_proto
		return init_proto(self.cfg, 'eth').checksummed_addr(addr)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

import sys, os
from pathlib import Path
from subprocess import run
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

def build_libsecp256k1():

	home          = Path(os.environ['HOME'])
	cache_path    = home.joinpath('.cache', 'mmgen')
	src_path      = home.joinpath('.cache', 'mmgen', 'secp256k1')
	lib_file      = src_path.joinpath('.libs', 'libsecp256k1.dll.a')
	root = Path().resolve().anchor
	installed_lib_file = Path(root, 'msys64', 'ucrt64', 'lib', 'libsecp256k1.dll.a')

	if installed_lib_file.exists():
		return

	# fix broken aclocal path:
	os.environ['ACLOCAL_PATH'] = '/C/msys64/ucrt64/share/aclocal:/C/msys64/usr/share/aclocal'

	if not src_path.exists():
		cache_path.mkdir(parents=True)
		print('\nCloning libsecp256k1')
		run(['git', 'clone', 'https://github.com/bitcoin-core/secp256k1.git'], check=True, cwd=cache_path)

	if not lib_file.exists():
		print(f'\nBuilding libsecp256k1 (cwd={str(src_path)})')
		cmds = (
			['sh', './autogen.sh'],
			['sh', './configure', 'CFLAGS=-g -O2 -fPIC', '--disable-dependency-tracking',
				'--enable-module-recovery', 'MAKE=mingw32-make'],
			['mingw32-make', 'MAKE=mingw32-make'])
		for cmd in cmds:
			print('Executing {}'.format(' '.join(cmd)))
			run(cmd, check=True, cwd=src_path)

	if not installed_lib_file.exists():
		run(['mingw32-make', 'install', 'MAKE=mingw32-make'], check=True, cwd=src_path)

class my_build_ext(build_ext):
	def build_extension(self, ext):
		if sys.platform == 'win32':
			build_libsecp256k1()
		build_ext.build_extension(self, ext)

setup(
	cmdclass = {'build_ext': my_build_ext},
	ext_modules = [Extension(
		name      = 'mmgen.proto.secp256k1.secp256k1',
		sources   = ['extmod/secp256k1mod.c'],
		depends   = ['extmod/random.h'],
		libraries = ['gmp', 'secp256k1', 'bcrypt'] if sys.platform == 'win32' else ['secp256k1'],
		include_dirs = ['/usr/local/include'] if sys.platform == 'darwin' else [],
		library_dirs = ['/usr/local/lib'] if sys.platform == 'darwin' else [],
	)]
)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.status: Bitcoin transaction status class
"""

import time

from ....tx import status as TxBase
from ....util import msg, suf
from ....util2 import format_elapsed_hr

class Status(TxBase.Status):

	async def display(self, *, idx=''):

		def do_return(exitval, message):
			if message:
				msg(message)
			return exitval

		assert idx == '', f'multiple txhex not supported for {self.tx.proto}'
		tx = self.tx

		class r:
			pass

		async def is_in_wallet():
			try:
				ret = await tx.rpc.icall(
					'gettransaction',
					txid              = tx.coin_txid,
					include_watchonly = True,
					verbose           = False)
			except:
				return False
			if ret.get('confirmations', 0) > 0:
				r.confs = ret['confirmations']
				return True
			else:
				return False

		async def is_in_utxos():
			try:
				return 'txid' in await tx.rpc.call('getrawtransaction', tx.coin_txid, True)
			except:
				return False

		async def is_in_mempool():
			try:
				await tx.rpc.call('getmempoolentry', tx.coin_txid)
				return True
			except:
				return False

		async def is_replaced():
			if await is_in_mempool():
				return False
			try:
				ret = await tx.rpc.icall(
					'gettransaction',
					txid              = tx.coin_txid,
					include_watchonly = True,
					verbose           = False)
			except:
				return False
			else:
				if 'bip125-replaceable' in ret and ret.get('confirmations', 1) <= 0:
					r.replacing_confs = -ret['confirmations']
					r.replacing_txs = ret['walletconflicts']
					return True
				else:
					return False

		if await is_in_mempool():
			d = await tx.rpc.icall(
				'gettransaction',
				txid              = tx.coin_txid,
				include_watchonly = True,
				verbose           = False)
			rep = ('' if d.get('bip125-replaceable') == 'yes' else 'NOT ') + 'replaceable'
			t = d['timereceived']
			if tx.cfg.quiet:
				msg('Transaction is in mempool')
			else:
				msg(f'TX status: in mempool, {rep}')
				msg('Sent {} ({})'.format(time.strftime('%c', time.gmtime(t)), format_elapsed_hr(t)))
			return do_return(0, '')
		elif await is_in_wallet():
			return do_return(0, f'Transaction has {r.confs} confirmation{suf(r.confs)}')
		elif await is_in_utxos():
			return do_return(4, 'ERROR: transaction is in the blockchain (but not in the tracking wallet)!')
		elif await is_replaced():
			msg('Transaction has been replaced')
			msg('Replacement transaction ' + (
					f'has {r.replacing_confs} confirmation{suf(r.replacing_confs)}'
				if r.replacing_confs else
					'is in mempool'))
			if not tx.cfg.quiet:
				msg('Replacing transactions:')
				d = []
				for txid in r.replacing_txs:
					try:
						d.append(await tx.rpc.call('getmempoolentry', txid))
					except:
						d.append({})
				for txid, mp_entry in zip(r.replacing_txs, d):
					msg(f'  {txid}' + (' in mempool' if 'height' in mp_entry else ''))
			return do_return(0, '')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
help.xmrwallet: xmrwallet help notes for the MMGen Wallet suite
"""

def help(proto, cfg):
	from ..xmrwallet import uarg_info

	return """
Many operations take an optional ‘wallets’ argument: one or more address
indexes (expressed as a comma-separated list and/or hyphenated range) in
the default or specified key-address file, each corresponding to a Monero
wallet with the same index.  If the argument is omitted, all wallets are
operated upon.

All operations except for ‘relay’ require a running Monero daemon (monerod).
Unless --daemon is specified, the daemon is assumed to be listening on
localhost at the default RPC port.

If --tx-relay-daemon is specified, the monerod at HOST:PORT will be used to
relay any created transactions.  PROXY_IP:PROXY_PORT, if specified, may point
to a SOCKS proxy, in which case HOST may be a Tor onion address.

All communications use the RPC protocol via SSL (HTTPS) or Tor.  RPC over
plain HTTP is not supported.


                            SUPPORTED OPERATIONS

create    - create wallets for all or specified addresses in key-address file
sync      - sync wallets for all or specified addresses in key-address file
            and display a summary of accounts and balances
list      - same as ‘sync’, but also list detailed address info for accounts
view      - display a summary of accounts and balances in offline mode.  May
            be invoked without a running monerod
listview  - same as ‘view’, but also list detailed address info for accounts
label     - set a label for an address
new       - create a new account in a wallet, or a new address in an account
transfer  - transfer specified XMR amount from specified wallet:account to
            specified address
sweep     - sweep funds in specified wallet:account to new address in same
            account, or new or specified account in another wallet
sweep_all - same as above, but sweep balances of all addresses in the account
relay     - relay a transaction from a transaction file created using ‘sweep’
            or ‘transfer’ with the --no-relay option
submit    - submit an autosigned transaction to a wallet and the network
resubmit  - resubmit most recently submitted autosigned transaction (other
            actions are required: see Exporting Outputs below)
abort     - abort the current transaction created with --autosign.  The
            transaction may be signed or unsigned
txview    - display detailed information about a transaction file or files
txlist    - same as above, but display terse information in tabular format
dump      - produce JSON dumps of wallet metadata (accounts, addresses and
            labels) for a list or range of wallets
restore   - same as ‘create’, but additionally restore wallet metadata from
            the corresponding JSON dump files created with ‘dump’
export-outputs      - export outputs of watch-only wallets for import into
                      their corresponding offline wallets
export-outputs-sign - same as above, plus request offline wallet to create
                      signed key images for import by ‘import-key-images’
import-key-images   - import key images signed by offline wallets into their
                      corresponding watch-only wallets


                           ‘LABEL’ OPERATION NOTES

This operation takes a LABEL_SPEC arg with the following format:

    {label_spec}

where WALLET is a wallet number, ACCOUNT an account index, and ADDRESS an
address index.


                            ‘NEW’ OPERATION NOTES

This operation takes a NEW_ADDRESS_SPEC arg with the following format:

    {newaddr_spec}

where WALLET is a wallet number and ACCOUNT an account index.  If ACCOUNT
is omitted, a new account will be created in the wallet.  Otherwise a new
address will be created in the specified account.  An optional label text
may be appended to the spec following a comma.


                         ‘TRANSFER’ OPERATION NOTES

The transfer operation takes a TRANSFER_SPEC arg with the following format:

    {transfer_spec}

where SOURCE is a wallet number, ACCOUNT the source account index, ADDRESS
the destination Monero address and AMOUNT the XMR amount to be sent.


                    ‘SWEEP’ AND ‘SWEEP_ALL’ OPERATION NOTES

The sweep and sweep_all operations take a SWEEP_SPEC arg with the following
format:

    {sweep_spec}

where SOURCE and DEST are wallet numbers and ACCOUNT account indices for the
respective wallets.

If DEST is omitted, a new address will be created in ACCOUNT of SOURCE, and
funds from ACCOUNT of SOURCE will be swept into it.

If DEST is included without its related ACCOUNT, funds from ACCOUNT of SOURCE
will be swept into a newly created account in DEST, or the last existing
account of DEST, if requested by the user.

If both account indices are included, funds from ACCOUNT of SOURCE will be
swept into ACCOUNT of DEST.

The user is prompted before addresses are created or funds transferred.

With ‘sweep’, if the source account has more than one address with a balance,
the balance of a single randomly chosen address will be swept.  To sweep the
balances of all addresses in an account, use ‘sweep_all’.


                    ‘SUBMIT’ AND ‘RELAY’ OPERATION NOTES

By default, transactions are relayed to a monerod running on localhost at the
default RPC port.  To relay transactions to a remote or non-default monerod
via optional SOCKS proxy, use the --tx-relay-daemon option described above.

When ‘submit’ is used with --autosign, the transaction filename must be
omitted.


                    ‘DUMP’ AND ‘RESTORE’ OPERATION NOTES

These commands produce and read JSON wallet dump files with the same
filenames as their source wallets, plus a .dump extension.

It’s highly advisable to make regular dumps of your Monero wallets and back
up the dump files, which can be used to easily regenerate the wallets using
the ‘restore’ operation, should the need arise.  For watch-only autosigning
wallets, creating the dumps is as easy as executing ‘mmgen-xmrwallet
--autosign dump’ from your wallet directory.  The dump files are formatted
JSON and thus suitable for efficient incremental backup using git.


                    ‘TXVIEW’ AND ‘TXLIST’ OPERATION NOTES

Transactions are displayed in chronological order based on submit time or
creation time.  With --autosign, submitted transactions on the removable
device are displayed.


                              SECURITY WARNING

If you have an existing MMGen Monero hot wallet setup, you’re strongly
advised to migrate to offline autosigning to avoid further exposing your
private keys on your network-connected machine.  See OFFLINE AUTOSIGNING
and ‘Replacing Existing Hot Wallets with Watch-Only Wallets’ below.


                                  EXAMPLES

Note that the transacting examples in this section apply for a hot wallet
setup, which is now deprecated.  See OFFLINE AUTOSIGNING below.

Generate an XMR key-address file with 5 addresses from your default wallet:
$ mmgen-keygen --coin=xmr 1-5

Create 3 Monero wallets from the key-address file:
$ mmgen-xmrwallet create *.akeys.mmenc 1-3

After updating the blockchain, sync wallets 1 and 2:
$ mmgen-xmrwallet sync *.akeys.mmenc 1,2

Sweep all funds from account #0 of wallet 1 to a new address:
$ mmgen-xmrwallet sweep *.akeys.mmenc 1:0

Same as above, but use a TX relay on the Tor network:
$ mmgen-xmrwallet --tx-relay-daemon=abcdefghijklmnop.onion:127.0.0.1:9050 sweep *.akeys.mmenc 1:0

Sweep all funds from account #0 of wallet 1 to wallet 2:
$ mmgen-xmrwallet sweep *.akeys.mmenc 1:0,2

Send 0.1 XMR from account #0 of wallet 2 to an external address:
$ mmgen-xmrwallet transfer *.akeys.mmenc 2:0:<monero address>,0.1

Sweep all funds from account #0 of wallet 2 to a new address, saving the
transaction to a file:
$ mmgen-xmrwallet --no-relay sweep *.akeys.mmenc 2:0

Relay the created sweep transaction via a host on the Tor network:
$ mmgen-xmrwallet --tx-relay-daemon=abcdefghijklmnop.onion:127.0.0.1:9050 relay *XMR*.sigtx

Create a new account in wallet 2:
$ mmgen-xmrwallet new *.akeys.mmenc 2

Create a new address in account 1 of wallet 2, with label:
$ mmgen-xmrwallet new *.akeys.mmenc 2:1,"from ABC exchange"

View all the XMR transaction files in the current directory, sending output
to pager:
$ mmgen-xmrwallet --pager txview *XMR*.sigtx


                             OFFLINE AUTOSIGNING

                                  Tutorial

Master the basic concepts of the MMGen wallet system and the processes of
wallet creation, conversion and backup described in the Getting Started
guide.  Optionally create a default MMGen wallet on your offline machine
using ‘mmgen-walletgen’.  If you choose not to do this, you’ll be prompted
for a seed phrase at the start of each signing session.

Familiarize yourself with the autosigning setup process as described in
‘mmgen-autosign --help’.  Prepare your removable device and set up the
mountpoints on your offline and online machines according to the instructions
therein.  Install ‘monero-wallet-rpc’ on your offline machine and the Monero
CLI wallet and daemon binaries on your online machine.

On the offline machine, insert the removable device and execute:

$ mmgen-autosign --xmrwallets=1-2,7 setup

This will create 3 Monero signing wallets with indexes 1, 2 and 7 and primary
addresses matching your seed’s Monero addresses with the same indexes.  (Note
that these particular indexes are arbitrary, for purposes of illustration
only.  Feel free to choose your own list and/or range – or perhaps just the
number ‘1’ if one wallet is all you require).

These signing wallets are written to volatile memory and exist only for the
duration of the signing session, just like the temporary MMGen signing wallet
they’re generated from (see ‘mmgen-autosign --help’).

A viewkey-address file for the 3 addresses will also be written to the
removable device.  The data in this file will be used to create and access
watch-only wallets on your online machine that match the signing wallets
you’ve just created.

When the setup operation completes, extract the removable device and restart
the autosign script in wait mode:

$ mmgen-autosign --coins=xmr --stealth-led wait

Your only further physical interaction with the offline signing machine now
(assuming everything goes as planned) will be inserting and extracting the
removable device on it.  And this is the whole point of autosigning: to make
cold signing as convenient as possible, almost like transacting with a hot
wallet.

If your signing machine is an SoC with MMGen Wallet LED support (see
‘mmgen-autosign --help’), a quickly flashing LED will indicate that signing
is in progress, a slowly flashing LED an error condition, and no LED that the
program is idle and waiting for device insertion.

On your online machine, start monerod, wait until it’s fully synced with the
network, insert the removable device and execute:

$ mmgen-xmrwallet --autosign --restore-height=current create

This will create 3 watch-only wallets matching your 3 offline signing wallets
and write them to the current directory (an alternate wallet directory may be
specified with the --wallet-dir option).

Note that --restore-height=current is required to prevent a time-consuming
full sync of the wallets from the Genesis block, a meaningless waste of time
in this case since the wallets contain no funds.

Also make note of the --autosign option, a requirement for ALL autosigning
operations with ‘mmgen-xmrwallet’.

Now list your newly created wallets:

$ mmgen-xmrwallet --autosign list

Note that you can also use the ‘sync’ operation here, which produces more
abbreviated output than ‘list’.

Send some XMR (preferably a tiny amount) to the primary address of wallet #7.
Once the transaction has confirmed, invoke ‘sync’ or ‘list’ again to verify
the funds have arrived.

Since offline wallet #7 has no knowledge of the funds received by its online
counterpart, we need to update its state.  Export the outputs of watch-only
wallet #7 as follows:

$ mmgen-xmrwallet --autosign export-outputs 7

The outputs are now saved to the removable device and will be imported into
offline wallet #7 when you sign your first transaction.

Now you’re ready to begin transacting.  Let’s start by sweeping your funds in
wallet #7’s primary address (account 0) to a new address in the same account:

$ mmgen-xmrwallet --autosign sweep 7:0

This operation creates an unsigned sweep transaction and saves it to the
removable device.

Now extract the removable device and insert it on the offline machine.  Wait
for the quick LED flashing to stop (or the blue ‘safe to extract’ message, in
the absence of LED support), signalling that signing is complete.

Note that the offline wallet has performed two operations in one go here:
an import of wallet outputs from the previous step and the signing of your
just-created sweep transaction.

Extract the removable device, insert it on your online machine and submit the
signed sweep transaction to the watch-only wallet, which will broadcast it to
the network:

$ mmgen-xmrwallet --autosign submit

Note that you may also relay the transaction to a remote daemon, optionally
via a Tor proxy, using the --tx-relay-daemon option documented above.

Once your transaction has confirmed, invoke ‘list’ or ‘sync’ to view your
wallets’ balances.

Congratulations, you’ve performed your first autosigned Monero transaction!

For other examples, consult the EXAMPLES section above, noting the following
differences that apply to autosigning:

  1) The --autosign option must always be included.
  2) The key-address file argument must always be omitted.
  3) The ‘relay’ operation is replaced by ‘submit’, with TX filename omitted.
  4) Always remember to sign your transactions after a ‘sweep’ or ‘transfer’
     operation.
  5) Always remember to export a wallet’s outputs when it has received funds
     from an outside source.


                              Exporting Outputs

Exporting outputs from a watch-only wallet is generally required in only
three cases:

  a) at the start of each signing session (after ‘mmgen-autosign setup’);
  b) after the wallet has received funds from an outside source or another
     wallet; and
  c) after performing a ‘resubmit’ operation.

You might also need to do it, however, if an offline wallet is unable to sign
a transaction due to missing outputs.

At the start of a new signing session, you must export outputs from ALL
wallets you intend to transact with.  This is necessary because the offline
signing wallets have just been created and know nothing about the state of
their online counterparts.

Export outputs from a wallet as follows (for all wallets, omit the index):

$ mmgen-xmrwallet --autosign export-outputs <wallet index>

Then insert the removable device on the offline machine.  This will import
the outputs into the corresponding signing wallet(s) (and optionally redo any
failed transaction signing operation).

Following a ‘resubmit’, use the ‘export-outputs-sign’ operation instead, and
add the --rescan-spent option:

$ mmgen-xmrwallet --autosign --rescan-spent export-outputs-sign <wallet index>

Here the offline signing wallet(s) will also create signed key images. Insert
the removable device on your online machine and import the signed key images
into your online wallet as follows:

$ mmgen-xmrwallet --autosign import-key-images

Usually, this is all that is required.  However, if your wallet continues to
show an incorrect balance after the import operation, you’ll need to re-run
‘export-outputs-sign’ with the --rescan-blockchain option, followed by another
offline signing and online key image import.  Note that blockchain rescans can
take a long time, so patience is required here.


           Replacing Existing Hot Wallets with Watch-Only Wallets

If you have an existing MMGen Monero hot wallet setup, you can migrate to
offline transaction signing by ‘cloning’ your existing hot wallets as
watch-only ones via the ‘dump’ and ‘restore’ operations described below.

For additional security, it’s also wise to create new watch-only wallets that
have never had keys exposed on an online machine and gradually transfer all
funds from your ‘cloned’ wallets to them.  The creation of new wallets is
explained in the Tutorial above.

Start the cloning process by making dump files of your hot wallets’ metadata
(accounts, subaddresses and labels).  ‘cd’ to the wallet directory (or use
--wallet-dir) and execute:

$ mmgen-xmrwallet dump /path/to/key-address-file.akeys{{.mmenc}}

If you’ve been transacting with the wallets, you know where their key-address
file is along with its encryption password, if any.  Supply an additional
index range and/or list at the end of the command line if the key-address
file contains more wallets than exist on disk or there are wallets you wish
to ignore.

Do a directory listing to verify that the dump files are present alongside
their source wallet files ending with ‘MoneroWallet’.  Then execute:

$ mmgen-xmrwallet --watch-only restore /path/to/key-address-file.akeys{{.mmenc}}

This will create watch-only wallets that “mirror” the old hot wallets and
populate them with the metadata saved in the dump files.

Note that watch-only wallet filenames end with ‘MoneroWatchOnlyWallet’.  Your
old hot wallets will be ignored from here on.  Eventually, you’ll want to
destroy them.

Your new wallets must now be synced with the blockchain.  Begin by starting
monerod and synchronizing with the network.

Mount ‘/mnt/mmgen_autosign’ and locate the file in the ‘xmr’ directory with
the .vkeys extension, which contains the passwords you’ll need to log into
the wallets.  This is a plain text file viewable with ‘cat’, ‘less’ or your
favorite text editor.

Then log into each watch-only wallet in turn as follows:

$ monero-wallet-cli --wallet <wallet filename>

Upon login, each wallet will begin syncing, a process which can take more
than an hour depending on your hardware.  Note, however, that the process
is interruptible: you may exit ‘monero-wallet-cli’ at any point, log back
in again and resume where you left off.

Once your watch-only wallets are synced, you need to export their outputs:

$ mmgen-xmrwallet --autosign export-outputs-sign

Now insert the removable device on the offline machine and wait until the LED
stops flashing (or ‘safe to extract’).  The wallet outputs are now imported
into the signing wallets and corresponding signed key images have been
written to the removable device.

Insert the removable device on your online machine and import the key images
into your watch-only wallets:

$ mmgen-xmrwallet --autosign import-key-images

Congratulations, your watch-only wallets are now complete and you may begin
transacting!  First perform a ‘sync’ or ‘list’ to ensure that your balances
are correct.  Then you might try sweeping some funds as described in the
Tutorial above.

Once you’ve gained proficiency with the autosigning process and feel ready
to delete your old hot wallets, make sure to do so securely using ‘shred’,
‘wipe’ or some other secure deletion utility.
""".strip().format(
	newaddr_spec  = uarg_info['newaddr_spec'].annot,
	label_spec    = uarg_info['label_spec'].annot,
	transfer_spec = uarg_info['transfer_spec'].annot,
	sweep_spec    = uarg_info['sweep_spec'].annot,
)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.brain: brainwallet wallet class
"""

from ..util import msg
from ..color import yellow
from .enc import wallet
from .seed import Seed

class wallet(wallet):

	stdin_ok = True
	desc = 'brainwallet'
	# brainwallet warning message? TODO

	def get_bw_params(self):
		# already checked
		a = self.cfg.brain_params.split(',')
		return int(a[0]), a[1]

	def _deformat(self):
		self.brainpasswd = ' '.join(self.fmt_data.split())
		return True

	def _decrypt(self):
		d = self.ssdata
		if self.cfg.brain_params:
			bw_seed_len, d.hash_preset = self.get_bw_params()
		else:
			if not self.cfg.seed_len:
				self.cfg._util.qmsg(f'Using default seed length of {yellow(str(Seed.dfl_len))} bits\n'
					+ 'If this is not what you want, use the --seed-len option')
			self._get_hash_preset()
			bw_seed_len = self.cfg.seed_len or Seed.dfl_len
		self.cfg._util.qmsg_r('Hashing brainwallet data.  Please wait...')
		# Use buflen arg of scrypt.hash() to get seed of desired length
		seed = self.crypto.scrypt_hash_passphrase(
			self.brainpasswd.encode(),
			b'',
			d.hash_preset,
			buflen = bw_seed_len // 8)
		self.cfg._util.qmsg('Done')
		self.seed = Seed(self.cfg, seed_bin=seed)
		msg(f'Seed ID: {self.seed.sid}')
		self.cfg._util.qmsg('Check this value against your records')
		return True

	def _format(self):
		raise NotImplementedError('Brainwallet not supported as an output format')

	def _encrypt(self):
		raise NotImplementedError('Brainwallet not supported as an output format')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.util: transaction utilities
"""

def get_autosign_obj(cfg):
	from ..cfg import Config
	from ..autosign import Autosign
	return Autosign(
		Config({
			'_clone': cfg,
			'mountpoint': cfg.autosign_mountpoint,
			'coins': cfg.coin,
			'online': not cfg.offline, # used only in online environment (xmrwallet, txcreate, txsend, txbump)
		}))

def mount_removable_device(cfg):
	asi = get_autosign_obj(cfg)
	if not asi.device_inserted:
		from ..util import die
		die(1, 'Removable device not present!')
	import atexit
	atexit.register(lambda: asi.do_umount())
	asi.do_mount()
	return asi



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
filename: File and MMGenFile classes and methods for the MMGen suite
"""

import sys, os
from .util import die, get_extension

class File:

	def __init__(self, fn, *, write=False):

		self.name     = fn
		self.dirname  = os.path.dirname(fn)
		self.basename = os.path.basename(fn)
		self.ext      = get_extension(fn)
		self.mtime    = None
		self.ctime    = None
		self.atime    = None

		try:
			st = os.stat(fn)
		except:
			die('FileNotFound', f'{fn!r}: file not found')

		import stat
		if stat.S_ISBLK(st.st_mode):
			if sys.platform in ('win32',):
				die(2, 'Access to raw block devices not supported on platform {sys.platform!r}')
			mode = (os.O_RDONLY, os.O_RDWR)[bool(write)]
			try:
				fd = os.open(fn, mode)
			except OSError as e:
				if e.errno == 13:
					die(2, f'{fn!r}: permission denied')
#				if e.errno != 17: raise
			else:
				if sys.platform == 'linux':
					self.size = os.lseek(fd, 0, os.SEEK_END)
				elif sys.platform == 'darwin':
					from .platform.darwin.util import get_device_size
					self.size = get_device_size(fn)
				os.close(fd)
		else:
			self.size  = st.st_size
			self.mtime = st.st_mtime
			self.ctime = st.st_ctime
			self.atime = st.st_atime

class FileList(list):

	def __init__(self, fns, *, write=False):
		list.__init__(
			self,
			[File(fn, write=write) for fn in fns])

	def names(self):
		return [f.name for f in self]

	def sort_by_age(self, *, key='mtime', reverse=False):
		assert key in ('atime', 'ctime', 'mtime'), f'{key!r}: invalid sort key'
		self.sort(key=lambda a: getattr(a, key), reverse=reverse)

class MMGenFile(File):

	def __init__(self, fn, *, base_class=None, subclass=None, proto=None, write=False):
		"""
		'base_class' - a base class with an 'ext_to_cls' method
		'subclass'   - a subclass with an 'ext' attribute

		One or the other must be provided, but not both.

		The base class signals support for the MMGenFile API by setting its 'filename_api'
		attribute to True.
		"""

		super().__init__(fn, write=write)

		assert (subclass or base_class) and not (subclass and base_class), 'MMGenFile chk1'

		if not getattr(subclass or base_class, 'filename_api', False):
			die(3, f'Class {(subclass or base_class).__name__!r} does not support the MMGenFile API')

		if base_class:
			subclass = base_class.ext_to_cls(self.ext, proto)
			if not subclass:
				die('BadFileExtension', f'{self.ext!r}: not a recognized file extension for {base_class}')

		self.subclass = subclass

class MMGenFileList(FileList):

	def __init__(self, fns, base_class, *, proto=None, write=False):
		list.__init__(
			self,
			[MMGenFile(fn, base_class=base_class, proto=proto, write=write) for fn in fns])

def find_files_in_dir(subclass, fdir, *, no_dups=False):

	assert isinstance(subclass, type), f'{subclass}: not a class'

	if not getattr(subclass, 'filename_api', False):
		die(3, f'Class {subclass.__name__!r} does not support the MMGenFile API')

	matches = [l for l in os.listdir(fdir) if l.endswith('.'+subclass.ext)]

	if no_dups:
		if len(matches) == 1:
			return os.path.join(fdir, matches[0])
		elif matches:
			die(1, f'ERROR: more than one {subclass.__name__} file in directory {fdir!r}')
		else:
			return None
	else:
		return [os.path.join(fdir, m) for m in matches]

def find_file_in_dir(subclass, fdir):
	return find_files_in_dir(subclass, fdir, no_dups=True)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
proto.eth.tw.unspent: Ethereum tracking wallet unspent outputs class
"""

from ....tw.unspent import TwUnspentOutputs
from .view import EthereumTwView

class EthereumTwUnspentOutputs(EthereumTwView, TwUnspentOutputs):

	hdr_lbl = 'tracked accounts'
	desc    = 'account balances'
	item_desc = 'account'
	item_desc_pl = 'accounts'

class EthereumTokenTwUnspentOutputs(EthereumTwUnspentOutputs):

	has_amt2 = True

	async def get_data(self, *args, **kwargs):
		await super().get_data(*args, **kwargs)
		for e in self.data:
			e.amt2 = await self.twctl.get_eth_balance(e.addr)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
amt: MMGen CoinAmt and related classes
"""

from decimal import Decimal
from .objmethods import Hilite, InitErrors
from .obj import get_obj

class CoinAmt(Decimal, Hilite, InitErrors): # abstract class
	"""
	Instantiating with 'from_decimal' rounds value down to 'max_prec' precision.
	For addition and subtraction, operand types must match.
	For multiplication and division, operand types may differ.
	Negative amounts, floor division and modulus operation are unimplemented.

	Decimal precision is set in init_proto()
	"""
	coin = 'Coin'
	color = 'yellow'

	max_prec = 0      # number of decimal places for this coin
	max_amt  = None   # coin supply if known, otherwise None
	units    = ()     # defined unit names, e.g. ('satoshi',...)

	def __new__(cls, num, *, from_unit=None, from_decimal=False):

		if isinstance(num, CoinAmt):
			raise TypeError(f'CoinAmt: {num} is instance of {cls.__name__}')

		try:
			if from_unit:
				assert from_unit in cls.units, f'{from_unit!r}: unrecognized coin unit for {cls.__name__}'
				assert isinstance(num, int), 'value is not an integer'
				me = Decimal.__new__(cls, num * getattr(cls, from_unit))
			elif from_decimal:
				assert isinstance(num, Decimal), f'number must be of type Decimal, not {type(num).__name__})'
				me = Decimal.__new__(cls, num.quantize(Decimal('10') ** -cls.max_prec))
			else:
				assert isinstance(num, str), f'non-string passed to {cls.__name__} initializer'
				me = Decimal.__new__(cls, num)
			assert me.normalize().as_tuple()[-1] >= -cls.max_prec, 'too many decimal places in coin amount'
			if cls.max_amt:
				assert me <= cls.max_amt, f'{me}: coin amount too large (>{cls.max_amt})'
			assert me >= 0, 'coin amount cannot be negative'
			return me
		except Exception as e:
			return cls.init_fail(e, num)

	def to_unit(self, unit):
		if (u := getattr(self, unit)) == self.atomic:
			return int(Decimal(self) // u)
		else:
			return Decimal('{:0.{w}f}'.format(
				self / u,
				w = (u/self.atomic).as_tuple().exponent))

	@classmethod
	def fmtc(cls, *args, **kwargs):
		cls.method_not_implemented()

	def fmt(self, iwidth=1, /, *, color=False, prec=None): # iwidth: width of the integer part
		prec = prec or self.max_prec
		if '.' in (s := str(self)):
			a, b = s.split('.', 1)
			return self.colorize(
				a.rjust(iwidth) + '.' + b.ljust(prec)[:prec], # truncation, not rounding!
				color = color)
		else:
			return self.colorize(
				s.rjust(iwidth).ljust(iwidth+prec+1),
				color = color)

	def hl(self, *, color=True):
		return self.colorize(str(self), color=color)

	# fancy highlighting with coin unit, enclosure, formatting
	def hl2(self, *, color=True, unit=False, fs='{}', encl=''):
		res = fs.format(self)
		return (
			encl[:-1]
			+ self.colorize(
				(res.rstrip('0').rstrip('.') if '.' in res else res) +
				(' ' + self.coin if unit else ''),
				color = color)
			+ encl[1:]
		)

	def __str__(self): # format simply, with no exponential notation
		return str(int(self)) if int(self) == self else self.normalize().__format__('f')

	def __repr__(self):
		return "{}('{}')".format(type(self).__name__, self.__str__())

	def __add__(self, other, *args, **kwargs):
		"""
		we must allow other to be int(0) to use the sum() builtin
		"""
		if type(other) is type(self) or (other == 0 and isinstance(other, int)):
			return type(self)(Decimal.__add__(self, other, *args, **kwargs),  from_decimal=True)
		raise TypeError(
			f'operand {other} is of incorrect type ({type(other).__name__} != {type(self).__name__})')

	__radd__ = __add__

	def __sub__(self, other, *args, **kwargs):
		if type(other) is type(self):
			return type(self)(Decimal.__sub__(self, other, *args, **kwargs),  from_decimal=True)
		raise TypeError(
			f'operand {other} is of incorrect type ({type(other).__name__} != {type(self).__name__})')

	def __rsub__(self, other, *args, **kwargs):
		if type(other) is type(self):
			return type(self)(Decimal.__rsub__(self, other, *args, **kwargs),  from_decimal=True)
		raise TypeError(
			f'operand {other} is of incorrect type ({type(other).__name__} != {type(self).__name__})')

	def __mul__(self, other, *args, **kwargs):
		return type(self)('{:0.{p}f}'.format(
			Decimal.__mul__(self, Decimal(other), *args, **kwargs),
			p = self.max_prec
		))

	__rmul__ = __mul__

	def __truediv__(self, other, *args, **kwargs):
		return type(self)('{:0.{p}f}'.format(
			Decimal.__truediv__(self, Decimal(other), *args, **kwargs),
			p = self.max_prec
		))

	def __rtruediv__(self, other, *args, **kwargs):
		return type(self)('{:0.{p}f}'.format(
			Decimal.__rtruediv__(self, Decimal(other), *args, **kwargs),
			p = self.max_prec
		))

	def __neg__(self, *args, **kwargs):
		self.method_not_implemented()

	def __floordiv__(self, *args, **kwargs):
		self.method_not_implemented()

	def __mod__(self, *args, **kwargs):
		self.method_not_implemented()

def is_coin_amt(proto, num, *, from_unit=None, from_decimal=False):
	assert proto.coin_amt, 'proto.coin_amt is None!  Did you call init_proto() with ‘need_amt’?'
	return get_obj(
		proto.coin_amt,
		num          = num,
		from_unit    = from_unit,
		from_decimal = from_decimal,
		silent       = True,
		return_bool  = True)

class BTCAmt(CoinAmt):
	coin = 'BTC'
	max_prec = 8
	max_amt = 21000000
	satoshi = Decimal('0.00000001')
	atomic = satoshi
	units = ('satoshi',)

class UniAmt(BTCAmt):
	coin = None

class BCHAmt(BTCAmt):
	coin = 'BCH'

class LTCAmt(BTCAmt):
	coin = 'LTC'
	max_amt = 84000000

class XMRAmt(CoinAmt):
	coin = 'XMR'
	max_prec = 12
	atomic = Decimal('0.000000000001')
	units = ('atomic',)

# Kwei (babbage) 3, Mwei (lovelace) 6, Gwei (shannon) 9, µETH (szabo) 12, mETH (finney) 15, ETH 18
class ETHAmt(CoinAmt):
	coin    = 'ETH'
	max_prec = 18
	wei     = Decimal('0.000000000000000001')
	Kwei    = Decimal('0.000000000000001')
	Mwei    = Decimal('0.000000000001')
	Gwei    = Decimal('0.000000001')
	szabo   = Decimal('0.000001')
	finney  = Decimal('0.001')
	atomic  = wei
	units   = ('wei', 'Kwei', 'Mwei', 'Gwei', 'szabo', 'finney')

	def toWei(self):
		return int(Decimal(self) // self.wei)

class ETCAmt(ETHAmt):
	coin = 'ETC'

class TokenAmt(ETHAmt):
	units = ('atomic',)

	def __new__(cls, num, *, decimals, from_unit=None):
		assert isinstance(decimals, int)
		cls.max_prec = decimals
		cls.atomic = Decimal(f'{10 ** -decimals:0.{decimals}f}')
		return ETHAmt.__new__(cls, num=num, from_unit=from_unit)

	def to_unit(self, unit):
		if (u := getattr(self, unit)) == self.atomic:
			return int(Decimal(self) // u)
		raise ValueError('TokenAmt unit must be ‘atomic’')

def CoinAmtChk(proto, num):
	assert type(num) is proto.coin_amt, f'CoinAmtChk: {type(num)} != {proto.coin_amt}'
	return num

class RelFeeAmt(Decimal):

	def __str__(self):
		return '{:.08f}'.format(self).rstrip('0').rstrip('.')



#!/usr/bin/env python3

from mmgen.cfg import Config
from mmgen.util import msg, oneshot_warning, oneshot_warning_group

cfg = Config()

class foo(oneshot_warning):

	color = 'purple'
	message = 'foo variant {} selected'

class bar(oneshot_warning):

	color = 'yellow'
	message = 'bar is experimental software'

class wg(oneshot_warning_group):

	class foo:
		color = 'yellow'
		message = 'foo is experimental software.  Proceed at your own risk'

	class bar:
		color = 'red'
		message = 'The bar command is {} and can break your {}!!!'

	class baz:
		color = 'orange'
		message = 'baz variant {} selected'

for i in (1, 2, 3):

	from mmgen.crypto import Crypto

	msg('\npw')
	for k in ('A', 'B'):
		ret = Crypto.pwfile_reuse_warning(k).warning_shown
		assert ret == (i != 1), 'warning_shown incorrect'

	msg('wg1')
	wg('foo')
	msg('wg2')
	wg('bar', fmt_args=['dangerous', 'computer'])
	msg('wg3')
	wg('baz', div='alpha', fmt_args=['alpha'])
	msg('wg4')
	wg('baz', div='beta', fmt_args=['beta'])

	msg('w1')
	foo(div='alpha', fmt_args=['alpha'])
	msg('w2')
	foo(div='beta', fmt_args=['beta'])
	msg('w3')
	bar()

	msg('bottom')



#!/usr/bin/env python3

"""
test.modtest_d.baseconv: Base conversion unit test for the MMGen suite
"""

from mmgen.util import msg, msg_r

from ..include.common import cfg, qmsg, qmsg_r, vmsg, vmsg_r

class unit_test:

	vectors = {
		'b58': (
			(('00', None),       '1'),
			(('00', 1),          '1'),
			(('00', 2),          '11'),
			(('01', None),       '2'),
			(('01', 1),          '2'),
			(('01', 2),          '12'),
			(('0f', None),       'G'),
			(('0f', 1),          'G'),
			(('0f', 2),          '1G'),
			(('deadbeef', None), '6h8cQN'),
			(('deadbeef', 20),   '111111111111116h8cQN'),
			(('00000000', None), '1'),
			(('00000000', 20),   '11111111111111111111'),
			(('ffffffff', None), '7YXq9G'),
			(('ffffffff', 20),   '111111111111117YXq9G'),
			(('ff'*16, 'seed'),  'YcVfxkQb6JRzqk5kF2tNLv'),
			(('ff'*24, 'seed'),  'QLbz7JHiBTspS962RLKV8GndWFwiEaqKL'),
			(('ff'*32, 'seed'),  'JEKNVnkbo3jma5nREBBJCDoXFVeKkD56V3xKrvRmWxFG'),
			(('00'*16, 'seed'),  '1111111111111111111111'),
			(('00'*24, 'seed'),  '111111111111111111111111111111111'),
			(('00'*32, 'seed'),  '11111111111111111111111111111111111111111111'),
		),
		# MMGen-flavored base32 using simple base conversion
		'b32': (
			(('00', None),       'A'),
			(('00', 1),          'A'),
			(('00', 2),          'AA'),
			(('01', None),       'B'),
			(('01', 1),          'B'),
			(('01', 2),          'AB'),
			(('0f', None),       'P'),
			(('0f', 1),          'P'),
			(('0f', 2),          'AP'),
			(('deadbeef', None), 'DPK3PXP'),
			(('deadbeef', 20),   'AAAAAAAAAAAAADPK3PXP'),
			(('00000000', None), 'A'),
			(('00000000', 20),   'AAAAAAAAAAAAAAAAAAAA'),
			(('ffffffff', None), 'D777777'),
			(('ffffffff', 20),   'AAAAAAAAAAAAAD777777'),
		),
		'b16': (
			(('00', None),       '0'),
			(('00', 1),          '0'),
			(('00', 2),          '00'),
			(('01', None),       '1'),
			(('01', 1),          '1'),
			(('01', 2),          '01'),
			(('0f', None),       'f'),
			(('0f', 1),          'f'),
			(('0f', 2),          '0f'),
			(('deadbeef', None), 'deadbeef'),
			(('deadbeef', 20),   '000000000000deadbeef'),
			(('00000000', None), '0'),
			(('00000000', 20),   '00000000000000000000'),
			(('ffffffff', None), 'ffffffff'),
			(('ffffffff', 20),   '000000000000ffffffff'),
		),
		'b10': (
			(('00', None),       '0'),
			(('00', 1),          '0'),
			(('00', 2),          '00'),
			(('01', None),       '1'),
			(('01', 1),          '1'),
			(('01', 2),          '01'),
			(('0f', None),       '15'),
			(('0f', 1),          '15'),
			(('0f', 2),          '15'),
			(('deadbeef', None), '3735928559'),
			(('deadbeef', 20),   '00000000003735928559'),
			(('00000000', None), '0'),
			(('00000000', 20),   '00000000000000000000'),
			(('ffffffff', None), '4294967295'),
			(('ffffffff', 20),   '00000000004294967295'),
		),
		'b8': (
			(('00', None),       '0'),
			(('00', 1),          '0'),
			(('00', 2),          '00'),
			(('01', None),       '1'),
			(('01', 1),          '1'),
			(('01', 2),          '01'),
			(('0f', None),       '17'),
			(('0f', 1),          '17'),
			(('0f', 2),          '17'),
			(('deadbeef', None), '33653337357'),
			(('deadbeef', 20),   '00000000033653337357'),
			(('00000000', None), '0'),
			(('00000000', 20),   '00000000000000000000'),
			(('ffffffff', None), '37777777777'),
			(('ffffffff', 20),   '00000000037777777777'),
		),
		'b6d': (
			(('00', None),       '1'),
			(('00', 1),          '1'),
			(('00', 2),          '11'),
			(('01', None),       '2'),
			(('01', 1),          '2'),
			(('01', 2),          '12'),
			(('0f', None),       '34'),
			(('0f', 1),          '34'),
			(('0f', 2),          '34'),
			(('010f', None),     '2242'),
			(('010f', 20),       '11111111111111112242'),
			(('deadbeef', None), '2525524636426'),
			(('deadbeef', 20),   '11111112525524636426'),
			(('00000000', None), '1'),
			(('00000000', 20),   '11111111111111111111'),
			(('ffffffff', None), '2661215126614'),
			(('ffffffff', 20),   '11111112661215126614'),

			(('ff'*16, 'seed'),  '34164464641266661652465154654653354436664555521414'),
			(('ff'*24, 'seed'),  '246111411433323364222465566552324652566121541623426135163525451613554313654'),
			(('ff'*32, 'seed'),  '2132521653312613134145131423465414636252114131225324246311141642456513416322412146151432142242565134'),
			(('00'*16, 'seed'),  '1'*50),
			(('00'*24, 'seed'),  '1'*75),
			(('00'*32, 'seed'),  '1'*100),
		),
		'mmgen': (
			(
				('deadbeefdeadbeefdeadbeefdeadbeef', 'seed'),
				'table cast forgive master funny gaze sadness ripple million paint moral match'
			), (
				('deadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef', 'seed'),
				'swirl maybe anymore mix scale stray fog use approach page crime rhyme ' +
				'class former strange window snap soon'
			), (
				('deadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef', 'seed'),
				'swell type milk figure cheese phone fill black test bloom heard comfort ' +
				'image terrible radio lesson own reply battle goal goodbye need laugh stream'
			), (
				('ffffffffffffffffffffffffffffffff', 'seed'),
				'yellow yeah show bowl season spider cling defeat poison law shelter reflect'
			), (
				('ffffffffffffffffffffffffffffffffffffffffffffffff', 'seed'),
				'yeah youth quit fail perhaps drum out person young click skin ' +
				'weird inside silently perfectly together anyone memory'
			), (
				('ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'seed'),
				'wrote affection object cell opinion here laughter stare honest north cost begin ' +
				'murder something yourself effort acid dot doubt game broke tell guilt innocent'
			), (
				('00000000000000000000000000000001', 'seed'),
				'able ' * 11 + 'about'
			), (
				('000000000000000000000000000000000000000000000001', 'seed'),
				'able ' * 17 + 'about'
			), (
				('0000000000000000000000000000000000000000000000000000000000000001', 'seed'),
				'able ' * 23 + 'about'
			),
		),
	}

	def run_test(self, name, ut):

		msg_r('Testing base conversion routines...')

		from mmgen.baseconv import baseconv
		perr = "length of {!r} less than pad length ({})"
		rerr = "return value ({!r}) does not match reference value ({!r})"

		qmsg_r('\nChecking hex-to-base conversion:')
		for base, data in self.vectors.items():
			fs = "  {h:%s}  {p:<6} {r}" % max(len(d[0][0]) for d in data)
			if not cfg.verbose:
				qmsg_r(f' {base}')
			vmsg(f'\nBase: {base}')
			vmsg(fs.format(h='Input', p='Pad', r='Output'))
			for (hexstr, pad), ret_chk in data:
				ret = baseconv(base).fromhex(hexstr, pad=pad, tostr=True)
				if pad != 'seed':
					assert len(ret) >= (pad or 0), perr.format(ret, pad or 0)
				assert ret == ret_chk, rerr.format(ret, ret_chk)
				vmsg(fs.format(h=hexstr, r=ret, p=str(pad)))

		qmsg_r('\nChecking base-to-hex conversion:')
		for base, data in self.vectors.items():
			fs = "  {h:%s}  {p:<6} {r}" % max(len(d[1]) for d in data)
			if not cfg.verbose:
				qmsg_r(f' {base}')
			vmsg(f'\nBase: {base}')
			vmsg(fs.format(h='Input', p='Pad', r='Output'))
			for (hexstr, pad), ret_chk in data:
				if type(pad) is int:
					pad = len(hexstr)
				ret = baseconv(base).tohex( ret_chk.split() if base == 'mmgen' else ret_chk, pad=pad )
				if pad is None:
					assert int(ret, 16) == int(hexstr, 16), rerr.format(int(ret, 16), int(hexstr, 16))
				else:
					assert ret == hexstr, rerr.format(ret, hexstr)
				vmsg(fs.format(h=ret_chk, r=ret, p=str(pad)))

		qmsg_r('\nChecking wordlist checksums:')
		vmsg('')

		for wl_id in baseconv.constants['wl_chksum']:
			vmsg_r(f'  {wl_id+":":9}')
			baseconv(wl_id).check_wordlist(cfg)

		qmsg('')

		qmsg('Checking error handling:')

		bad_b58 = 'I'*22
		bad_b58len = 'a'*23

		fr58 = baseconv('b58').fromhex
		to58 = baseconv('b58').tohex
		to32 = baseconv('b32').tohex
		to8  = baseconv('b8').tohex

		hse = 'HexadecimalStringError'
		bce = 'BaseConversionError'
		bpe = 'BaseConversionPadError'
		sle = 'SeedLengthError'

		bad_data = (
			('hexstr',           hse, ': not a hexadecimal str', lambda: fr58('x')),
			('hexstr (seed)',    hse, 'seed data not a hexadec', lambda: fr58('x', pad='seed')),
			('hexstr (empty)',   bce, 'empty data not allowed',  lambda: fr58('')),
			('b58 data',         bce, ': not in base58',         lambda: to58('IfFzZ')),
			('b58 data (seed)',  bce, 'seed data not in base58', lambda: to58(bad_b58, pad='seed')),
			('b58 len (seed)',   bce, 'invalid length for',      lambda: to58(bad_b58len, pad='seed')),
			('b58 data (empty)', bce, 'empty base58 data',       lambda: to58('')),
			('b8 data (empty)' , bce, 'empty base8 string data', lambda: to8('')),
			('b32 data',         bce, 'not in MMGen base32',     lambda: to32('1az')),
			('pad arg (in)',     bpe, "illegal value for 'pad'", lambda: fr58('ff', pad='foo')),
			('pad arg (in)',     bpe, "illegal value for 'pad'", lambda: fr58('ff', pad=False)),
			('pad arg (in)',     bpe, "illegal value for 'pad'", lambda: fr58('ff', pad=True)),
			('seedlen (in)',     sle, 'invalid byte length',     lambda: fr58('ff', pad='seed')),
			('pad arg (out)',    bpe, "illegal value for 'pad'", lambda: to58('Z', pad='foo')),
			('pad arg (out)',    bpe, "illegal value for 'pad'", lambda: to58('Z', pad=False)),
			('pad arg (out)',    bpe, "illegal value for 'pad'", lambda: to58('Z', pad=True)),
			('seedlen (out)',    bce, 'invalid length for seed', lambda: to58('Z', pad='seed')),
		)

		ut.process_bad_data(bad_data)

		msg('OK')

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
rpc.backends.requests: requests RPC backend for the MMGen Project
"""

import json

from ..util import dmsg_rpc_backend, json_encoder

from .base import base

class requests(base):

	def __del__(self):
		self.session.close()

	def __init__(self, caller):
		super().__init__(caller)
		import requests, urllib3
		urllib3.disable_warnings()
		self.session = requests.Session()
		self.session.trust_env = False # ignore *_PROXY environment vars
		self.session.headers = caller.http_hdrs
		if caller.auth_type:
			auth = 'HTTP' + caller.auth_type.capitalize() + 'Auth'
			self.session.auth = getattr(requests.auth, auth)(*caller.auth)
		if self.proxy: # used only by XMR for now: requires pysocks package
			self.session.proxies.update({
				'http':  f'socks5h://{self.proxy}',
				'https': f'socks5h://{self.proxy}'
			})

	async def run(self, *args, **kwargs):
		return self.run_noasync(*args, **kwargs)

	def run_noasync(self, payload, timeout, host_path):
		dmsg_rpc_backend(self.host_url, host_path, payload)
		res = self.session.post(
			url     = self.host_url + host_path,
			data    = json.dumps(payload, cls=json_encoder),
			timeout = timeout or self.timeout,
			verify  = False)
		return (res.content, res.status_code)
